﻿CompilerIf Defined(INCLUDE_SIMPLESTACKCLASS, #PB_Constant)=0
#INCLUDE_SIMPLESTACKCLASS=1
;/////////////////////////////////////////////////////////////////////////////////
;An OOP class for implementing a simple integer stack.

;© Big Eared Software 2016.
;==========================
;Updated 2016.

;Developed with Purebasic 5.42.
;Fully cross-platform.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;To create an instance of a new stack class, use something like :
;
;   MyStackObject.IntegerStackObject
;   MyStackObject = NewIntegerStack(numberOfElements)
;
;and check the return value is non zero.
;/////////////////////////////////////////////////////////////////////////////////
;/////////////////////////////////////////////////////////////////////////////////
;Methods exposed by the simple integer stack class :
;
;      Destroy()
;      NumberOfElementsPushed.i()
;      Peek.i(index)
;      Pop.i()
;      Push.i(value)
;      ResetStack()
;      ResizeStack.i(NumElements)
;      SizeOfStack.i()
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-Constants.

  ;Use the following in the \Peek() method.
    #INTEGERSTACK_GETLASTELEMENT           = -1
    #INTEGERSTACK_GETPOINTERTOLASTELEMENT  = -2
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-Class interfaces.

  ;The following interface contains the class methods.
    Interface IntegerStackObject
      Destroy()
      NumberOfElementsPushed.i()
      Peek.i(index=#INTEGERSTACK_GETLASTELEMENT)   
      Pop.i()
      Push.i(value)
      ResetStack()
      ResizeStack.i(NumElements)
      SizeOfStack.i()
    EndInterface 
;/////////////////////////////////////////////////////////////////////////////////


;The following structure contains the class template and private properties,
  Structure _membersIntegerStackClass
    *vTable
    NumElements.i
    StackBase.i
    StackPtr.i
  EndStructure 


;/////////////////////////////////////////////////////////////////////////////////
;-Create an instance of the integer stack class.
;Returns zero if an error.
Procedure.i NewIntegerStack(NumElements)
  Protected *object._membersIntegerStackClass, StackBase
  If NumElements > 0
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersIntegerStackClass))
      If *object
        ;Allocate memory for the stack.
          StackBase = AllocateMemory(NumElements*SizeOf(INTEGER))
          If stackBase
            ;Set the object properties.
              *object\vTable = ?VTable_IntegerStackClass
              *object\NumElements = numElements
              *object\StackBase= stackBase
              *object\StackPtr = 0
          Else
            FreeMemory(*object)
            *object = 0          
          EndIf
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;-CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following method frees all memory associated with the underlying stack.
Procedure stackClass_Destroy(*this._membersIntegerStackClass)
  FreeMemory(*this\StackBase)
  FreeMemory(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the number of elements pushed onto the stack.
Procedure.i stackBaseClass_NumberOfElementsPushed(*this._membersIntegerStackClass)
  ProcedureReturn *this\StackPtr
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the specified element from the underlying stack without
;performing any kind of 'pop'.
;'Index' = 0 or 1 or 2 or ... or NumberOfElementsPushed()-1.
Procedure.i stackClass_Peek(*this._membersIntegerStackClass, index=#INTEGERSTACK_GETLASTELEMENT)
  Protected *ptr.INTEGER, t1
  If index = #INTEGERSTACK_GETLASTELEMENT Or index = #INTEGERSTACK_GETPOINTERTOLASTELEMENT
    t1 = *this\StackPtr - 1
  Else
    t1 = index
  EndIf
  If t1 >= 0 And t1 < *this\StackPtr
  ;Good to go.
    *ptr = *this\StackBase + t1*SizeOf(INTEGER)
    If index = #INTEGERSTACK_GETPOINTERTOLASTELEMENT
      ProcedureReturn *ptr
    Else
      ProcedureReturn *ptr\i
    EndIf
EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following pops an element from the underlying stack.
Procedure.i stackClass_Pop(*this._membersIntegerStackClass)
  Protected *ptr.INTEGER
  If *this\StackPtr
    *this\StackPtr - 1
    *ptr = *this\StackBase + *this\StackPtr*SizeOf(INTEGER)
    ProcedureReturn *ptr\i
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following pushes a new element to the underlying stack if there is room.
;Returns zero if an error, i.e. if the stack is full.
Procedure.i stackClass_Push(*this._membersIntegerStackClass, Value)
  Protected *ptr.INTEGER
  If *this\StackPtr < *this\NumElements
    *ptr = *this\StackBase + *this\StackPtr*SizeOf(INTEGER)
    *ptr\i = Value
    *this\StackPtr + 1
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following resets a stack and effectively clears it's contents.
Procedure stackClass_ResetStack(*this._membersIntegerStackClass)
  *this\StackPtr = 0
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following attempts to dynamically resize the underlying stack.
;If the stack is reduced in size then the stack pointer may be lowered in order
;to point at the top of the stack etc.
;Returns zero if an error.
Procedure.i stackClass_ResizeStack(*this._membersIntegerStackClass, numElements)
  Protected mem
  If numElements And *this\NumElements <> numElements
    mem = ReAllocateMemory(*this\StackBase, numElements*SizeOf(INTEGER))
    If mem ;Reallocation was successful.
      *this\NumElements = numElements
      *this\StackBase = mem
      ;Reset the stack pointer if required.
        If *this\StackPtr > numElements
          *this\StackPtr = numElements
        EndIf
    EndIf
  EndIf
  ProcedureReturn mem
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the size of the stack.
Procedure.i stackBaseClass_SizeOfStack(*this._membersIntegerStackClass)
  ProcedureReturn *this\NumElements
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VIRTUAL TABLES.

DataSection 
  VTable_IntegerStackClass: 
    Data.i @stackClass_Destroy() 
    Data.i @stackBaseClass_NumberOfElementsPushed() 
    Data.i @stackClass_Peek() 
    Data.i @stackClass_Pop() 
    Data.i @stackClass_Push() 
    Data.i @stackClass_ResetStack() 
    Data.i @stackClass_ResizeStack() 
    Data.i @stackBaseClass_SizeOfStack() 
EndDataSection 
CompilerEndIf

; IDE Options = PureBasic 5.60 (Windows - x64)
; CursorPosition = 103
; FirstLine = 102
; Folding = --