﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Arrays module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;     i)    minScript arrays are user defined types. They are reference counted (not all user defined types need use reference counting).
;
;     ii)   An array can be created both within and outside of a minScript script. There are functions within this module which can be used
;           to create and maintain a minScript array.
;
;           In script use statements such as :
;             DIM myArray as ARRAY
;             LET myArray = NEWARRAY(#BYTE, 10, 20) or LET myArray = NEWARRAY("byte", 10, 20) etc.
;
;           to initialise a two-dimensional byte array. (Max num dimensions = 5).
;
;           NOTE that only script created arrays are covered by our array GC extension. That is to say that all other arrays will never be forcibly disposed of.
;
;NOTE to self. Variant arrays have \arrayType set to #minScript_VARIANT, but individual variant elements do not have this bit set.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES

  Declare.i minScript_RunTime_NewArray(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariant, *ret.minScriptVariant)
  Declare minScriptArray_GCE(codeObject, uMsg)
  Declare minScriptObjectVariantClass_Array_Clear(*this)
  Declare.i minScript_VarInitArray(*var.minScriptVariant, arrayType.u=#minScript_INTEGER)
  Declare.i minScript_VarSetArray(*var.minScriptVariant, upperBound0=-1, upperBound1=-1, upperBound2=-1, upperBound3=-1, upperBound4=-1)
  Declare.i minScript_VarGetArrayElement(*var.minScriptVariant, index0, index1=-1, index2=-1, index3=-1, index4=-1)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

  ;If changing the following then you will need to alter some of the functions within this module by hand.
    #minScript_MAXARRAYDIMS = 5 
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES

  ;The following is the class template for the minScriptObjectVariant class representing our arrays.
    Structure st_minScriptObjectVariantClass_Array
      *vTable
      refCount.i                        ;When this hits zero, the object is removed.
      ;Array fields.
        arrayType.u                     ;E.g. #minScript_BYTE
        vtConversionType.u              ;E.g. #minScript_INTEGER. For convenience.
        numDims.i
        numElements.i                   ;Total number of array elements (not bytes).
        sizeOfElement.i                 ;For convenience.
        arrayUpperBounds.l[#minScript_MAXARRAYDIMS]
        *pData.st_minScript_ArrayData   ;A separate memory allocation is required because multiple variants
                                        ;can point to the same array and we thus require memory reallocations to not alter the address of the st_minScriptObjectVariantClass_Array structure.
      ;The following safeguard is probably not needed, but is intended to prevent the object being destroyed whilst it is being destroyed through a circular reference and some unforseen action that might increase the reference count etc.
        blnBeingDestroyed.i
    EndStructure

  ;The following structure allows quick access to array data.
    Structure st_minScript_ArrayData
      StructureUnion
        b.b[0]
        w.w[0]
        l.l[0]
        i.i[0]
        q.q[0]
        f.f[0]
        d.d[0]
        r.d[0]
        sPtr.i[0]
        refPtr.minScriptObjectVariant[0]
        variant.minScriptVariant[0]
      EndStructureUnion
    EndStructure

  ;The following holds info on individual arrays so that our GC extension can dispose of them if needs be.
    Structure st_minScriptArray_GCE_ArrayEntry
      codeObject.i
      *refPtr.st_minScriptObjectVariantClass_Array
    EndStructure

  ;The following holds the information required to administer the class GC extension.
    Structure st_minScriptArray_GCE
      threadMutex.i
      Map arrayObjects.st_minScriptArray_GCE_ArrayEntry()  
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS

  Global minScriptArrayGCE_globals.st_minScriptArray_GCE
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following array init function is called at startup to allow minScript to register the array types etc.
;A host app would use similar code to register it's own user defined types.
Procedure.i minScript_InitArrayTypes()
  Protected result = #minScript_OKAY
  ;Register the array type.
    result = minScript_AddUserDefinedType("Array", #minScript_ARRAY)
  If result = #minScript_OKAY
    ;Register a run-time function used to create new instances of array objects.
    ;This particular function is required to return an instance of a minScriptObjectVariant interface.
      result = minScript_AddRuntimeFunction("NEWARRAY", #minScript_STANDARDRUNTIME_NEWARRAY, @minScript_RunTime_NewArray())
    If result = #minScript_OKAY
      ;Create a mutex to protect access to the GCE data if appropriate.
        CompilerIf #PB_Compiler_Thread
          minScriptArrayGCE_globals\threadMutex = CreateMutex()
          If minScriptArrayGCE_globals\threadMutex = 0
            result = #minScript_OUTOFMEMORY
          EndIf
        CompilerEndIf
      If result = #minScript_OKAY
        ;Add appropriate GC extensions.
          minScript_AddGCExtension(@minScriptArray_GCE())
      EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our "NEWARRAY" runtime function used to initialise an array object.
Procedure.i minScript_RunTime_NewArray(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, i, upperBound0, upperBound1=-1, upperBound2=-1, upperBound3=-1, upperBound4=-1, arrayType.u, *ptrString.STRING
  If numParamsPassed >= 2 And numParamsPassed <= 6
    ;The first parameter represents the array type which can be given in integer (in the form of a VT value) or string form.
    ;If in integer form then we cannot reliably check the array type as it can be a user-defined type.
    If *params\index[0]\vt = #minScript_STRING
      *ptrString = *params\index[0]
      arrayType = minScriptCodeClass_HELPER_String2VT(codeObject, LCase(*ptrString\s))
      If arrayType = 0
        ;We can try "byte" etc.
        Select LCase(*ptrString\s)
          Case "byte"
            arrayType = #minScript_BYTE
          Case "word"
            arrayType = #minScript_WORD
          Case "character"
            arrayType = #minScript_CHARACTER
          Case "long"
            arrayType = #minScript_LONG
          Case "float"
            arrayType = #minScript_FLOAT
        EndSelect
      EndIf
    ElseIf *params\index[0]\vt = #minScript_INTEGER   
      arrayType = *params\index[0]\i
      ;Check if the VT type is a registered type.
        If minScriptCodeClass_DISASSEMBLY_VT2String(arrayType, codeObject) = ""
          arrayType = 0
        EndIf
    EndIf
    ;We cannot reliably check the array type (parameter 0) as it can be a user-defined type. We can check it's type however.
    If arrayType
      ;We can check the parameter bounds and change their types where possible.
      For i = 1 To numParamsPassed - 1
        If *params\index[i]\vt & #minScript_REFERENCE
          result = #minScript_INVALIDDATATYPE
          Break
        ElseIf *params\index[i]\vt <> #minScript_INTEGER
          result = minScript_VarChangeType(@*params\index[i], #minScript_INTEGER)
          If result <> #minScript_OKAY
            result = #minScript_INVALIDDATATYPE ;For consistency.
            Break
          EndIf
        EndIf
        Select i-1
          Case 0
            upperBound0 = *params\index[i]\i
          Case 1
            upperBound1 = *params\index[i]\i
          Case 2
            upperBound2 = *params\index[i]\i
          Case 3
            upperBound3 = *params\index[i]\i
          Case 4
            upperBound4 = *params\index[i]\i
        EndSelect
      Next
      If result = #minScript_OKAY
        result = minScript_VarInitArray(*ret, arrayType)
        If result = #minScript_OKAY
          result = minScript_VarSetArray(*ret, upperBound0, upperBound1, upperBound2, upperBound3, upperBound4)
          If result = #minScript_OKAY
            ;Add to the array GC extension.
              CompilerIf #PB_Compiler_Thread
                LockMutex(minScriptArrayGCE_globals\threadMutex)
              CompilerEndIf
                If AddMapElement(minScriptArrayGCE_globals\arrayObjects(), Str(*ret\refPtr))
                  minScriptArrayGCE_globals\arrayObjects()\codeObject = codeObject
                  minScriptArrayGCE_globals\arrayObjects()\refPtr = *ret\refPtr
                Else
                  minScriptObjectVariantClass_Array_Clear(*ret\refPtr)
                  *ret\refPtr = 0
                  *ret\vt = #minScript_EMPTY
                  result = #minScript_OUTOFMEMORY
                EndIf
              CompilerIf #PB_Compiler_Thread
                UnlockMutex(minScriptArrayGCE_globals\threadMutex)
              CompilerEndIf
          Else
            minScriptObjectVariantClass_Array_Clear(*ret\refPtr)
            *ret\refPtr = 0
            *ret\vt = #minScript_EMPTY
          EndIf
        EndIf
      EndIf
    Else
      result = #minScript_INVALIDDATATYPE
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS  
  EndIf
  ;We will clear #minScript_OUTOFMEMORY errors so that scripts can determine for themselves how to proceed. A script can always use the THROWERROR() run-time function.
  If result = #minScript_OUTOFMEMORY
    result = #minScript_OKAY
  EndIf  
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-GC Extension.

;/////////////////////////////////////////////////////////////////////////////////
;The following GC extension disposes of any arrays unaccounted for.
Procedure minScriptArray_GCE(codeObject, uMsg)
  Protected i
  Select uMsg
    Case #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED
      ;Not needed.
    Default ; #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET or #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED
      ;We simply loop through all outstanding arrays and free those belonging to the given code object.
        CompilerIf #PB_Compiler_Thread
          LockMutex(minScriptArrayGCE_globals\threadMutex)
        CompilerEndIf
          ForEach minScriptArrayGCE_globals\arrayObjects()
            If minScriptArrayGCE_globals\arrayObjects()\codeObject = codeObject
              ;We forcibly delete the array object by clearing all elements.
              ;We must avoid clearing any reference type elements which only leaves string and variant type arrays requiring emptying.
                If minScriptArrayGCE_globals\arrayObjects()\refPtr\pData
                  If minScriptArrayGCE_globals\arrayObjects()\refPtr\arrayType = #minScript_VARIANT
                    For i = 0 To minScriptArrayGCE_globals\arrayObjects()\refPtr\numElements-1
                      If minScriptArrayGCE_globals\arrayObjects()\refPtr\pData\variant[i]\vt = #minScript_STRING
                        minScript_StringFree(minScriptArrayGCE_globals\arrayObjects()\refPtr\pData\variant[i]\sPtr)
                      EndIf
                    Next
                  ElseIf minScriptArrayGCE_globals\arrayObjects()\refPtr\arrayType = #minScript_STRING
                    For i = 0 To minScriptArrayGCE_globals\arrayObjects()\refPtr\numElements-1
                      minScript_StringFree(minScriptArrayGCE_globals\arrayObjects()\refPtr\pData\sPtr[i])
                    Next
                  EndIf
                  FreeMemory(minScriptArrayGCE_globals\arrayObjects()\refPtr\pData)
                EndIf
              FreeStructure(minScriptArrayGCE_globals\arrayObjects()\refPtr)
              DeleteMapElement(minScriptArrayGCE_globals\arrayObjects())
            EndIf
          Next
        CompilerIf #PB_Compiler_Thread
          UnlockMutex(minScriptArrayGCE_globals\threadMutex)
        CompilerEndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-minScriptObjectVariant CLASS METHODS.
;These 3 methods are required for scripts to be able to make use of our arrays.

;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \Clear() method.
;When clearing an array of objects, we only decrement the reference counts of the individual elements if the underlying array is to be
;completely removed. 
;We return only the first error code detected (which can only be a runtime error generated by a class destructor etc.)
Procedure minScriptObjectVariantClass_Array_Clear(*this.st_minScriptObjectVariantClass_Array)
  Protected result = #minScript_OKAY, i, t1
  *this\refCount - 1
  If *this\refCount = 0 And *this\blnBeingDestroyed = #False
    *this\blnBeingDestroyed = #True
    ;Clear the array data if present.
    If *this\pData
      If *this\arrayType = #minScript_VARIANT
        For i = 0 To *this\numElements-1
          If *this\pData\variant[i]\vt & #minScript_REFERENCE
            If *this\pData\variant[i]\refPtr
              t1 = *this\pData\variant[i]\refPtr\Clear()
              If result = #minScript_OKAY
                result = t1
              EndIf
              *this\pData\variant[i]\refPtr = 0 ;In case some wayward class destructor increases the ref count of the array.
            EndIf
          ElseIf *this\pData\variant[i]\vt = #minScript_STRING
            minScript_StringFree(*this\pData\variant[i]\sPtr)
            *this\pData\variant[i]\sPtr = 0 ;In case some wayward class destructor increases the ref count of the array.
          EndIf
        Next
      ElseIf *this\arrayType & #minScript_REFERENCE
        For i = 0 To *this\numElements-1
          If *this\pData\refPtr[i] 
            t1 = *this\pData\refPtr[i]\Clear()
            If result = #minScript_OKAY
              result = t1
            EndIf
            *this\pData\refPtr[i] = 0 ;In case some wayward class destructor increases the ref count of the array.
          EndIf
        Next
      ElseIf *this\arrayType = #minScript_STRING
        For i = 0 To *this\numElements-1
          minScript_StringFree(*this\pData\sPtr[i])
          *this\pData\sPtr[i] = 0 ;In case some wayward class destructor increases the ref count of the array.
        Next
      EndIf
      If *this\refCount = 0 ;In case some wayward class destructor increases the ref count of the array.
        FreeMemory(*this\pData)
      EndIf
    EndIf
    If *this\refCount = 0 ;In case some wayward class destructor increases the ref count of the array.
      FreeStructure(*this)
      ;Remove from the array GC extension.
        CompilerIf #PB_Compiler_Thread
          LockMutex(minScriptArrayGCE_globals\threadMutex)
        CompilerEndIf
          DeleteMapElement(minScriptArrayGCE_globals\arrayObjects(), Str(*this))
        CompilerIf #PB_Compiler_Thread
          UnlockMutex(minScriptArrayGCE_globals\threadMutex)
        CompilerEndIf
    Else
      *this\blnBeingDestroyed = #False    
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \CopyToVariant() method.
;When copying an array of objects, we do not increment the reference counts of the individual elements. 
;This is not required and could actually lead to memory leaks in the case that such an array is resized or individual elements overwritten etc.
Procedure.i minScriptObjectVariantClass_Array_CopyToVariant(*this.st_minScriptObjectVariantClass_Array, *varDest.minScriptVariant)
  Protected result = #minScript_OKAY, i
  *varDest\vt = #minScript_ARRAY ;Just in case!
  *varDest\refPtr = *this
  *this\refCount + 1
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Any underlying array will not only be correctly initialised, but will have a valid pData pointer.
Procedure.i minScriptObjectVariantClass_Array_Invoke(*this.st_minScriptObjectVariantClass_Array, codeObject.minScriptCode, methodName$, invokeType, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, i, index0, index1=-1, index2=-1, index3=-1, index4=-1, *ptrValue.st_minScript_memoryPointer
  Protected var.minScriptVariant, type$, vType, blnGetAddress
  If Left(methodName$, 1) = "@"
    blnGetAddress = #True
    methodName$ = Right(methodName$, Len(methodName$)-1)
  EndIf
  Select methodName$
    Case #minScript_DEFAULTPROPERTY$ ;Either a get, a set or a @get.
      If invokeType&#minScript_METHOD_GET Or blnGetAddress
        If numParamsPassed >= 2 And numParamsPassed <= 6
          ;Sort out the indexes and convert to integer if required.
          For i = 1 To numParamsPassed - 1
            If *params\index[i]\vt & #minScript_REFERENCE Or *params\index[i]\vt = #minScript_STRING
              result = #minScript_INVALIDORMISSINGARRAYINDEX
              Break
            ElseIf *params\index[i]\vt <> #minScript_INTEGER
              result = minScript_VarChangeType(@*params\index[i], #minScript_INTEGER)
              If result <> #minScript_OKAY
                result = #minScript_INVALIDORMISSINGARRAYINDEX ;For consistency.
                Break
              EndIf
            EndIf
            Select i-1
              Case 0
                index0 = *params\index[i]\i
              Case 1
                index1 = *params\index[i]\i
              Case 2
                index2 = *params\index[i]\i
              Case 3
                index3 = *params\index[i]\i
              Case 4
                index4 = *params\index[i]\i
            EndSelect
          Next
          If result = #minScript_OKAY
            *ptrValue = minScript_VarGetArrayElement(*params\index[0], index0, index1, index2, index3, index4)
            If *ptrValue
              If blnGetAddress
                *ret\vt = #minScript_INTEGER
                If *this\arrayType = #minScript_STRING Or (*this\arrayType = #minScript_VARIANT And *ptrValue\variant\vt = #minScript_STRING)
                  *ret\i = *ptrValue\i
                Else
                  *ret\i = *ptrValue
                EndIf
              Else
                ;Now we copy the value to the *ret variant.
                  If *this\arrayType = #minScript_VARIANT
                    *ret\vt = *ptrValue\variant\vt
                    vType = *ret\vt
                  Else
                    *ret\vt = *this\vtConversionType
                    vType = *this\arrayType
                  EndIf
                  If vType & #minScript_REFERENCE
                    If *ptrValue\refPtr
                      result = *ptrValue\refPtr\CopyToVariant(*ret)
                    Else
                      *ret\q = 0
                    EndIf
                  Else
                    Select vType
                      Case #minScript_EMPTY
                        *ret\i = 0
                      Case #minScript_BYTE
                        *ret\i = *ptrValue\b
                      Case #minScript_WORD
                        *ret\i = *ptrValue\w
                      Case #minScript_CHARACTER
                        *ret\i = *ptrValue\c
                      Case #minScript_LONG
                        *ret\i = *ptrValue\l
                      Case #minScript_FLOAT
                        *ret\r = *ptrValue\f
                      Case #minScript_INTEGER
                        *ret\i = *ptrValue\i
                      Case #minScript_CURRENCY
                        *ret\c = *ptrValue\q
                      Case #minScript_REAL
                        *ret\r = *ptrValue\d
                      Case #minScript_STRING
                        *ret\sPtr = minScript_StringAlloc(*ptrValue\i)
                    EndSelect          
                  EndIf
              EndIf
            Else
              result = #minScript_INVALIDORMISSINGARRAYINDEX
            EndIf
          EndIf
        Else
          result = #minScript_INVALIDORMISSINGARRAYINDEX
        EndIf
      Else ;Set default property.
        If *this\blnBeingDestroyed = #False ;We don't return an error here.
          If numParamsPassed >= 3 And numParamsPassed <= 7
            For i = 1 To numParamsPassed - 2
              If *params\index[i]\vt & #minScript_REFERENCE Or *params\index[i]\vt = #minScript_STRING
                result = #minScript_INVALIDORMISSINGARRAYINDEX
                Break
              ElseIf *params\index[i]\vt <> #minScript_INTEGER
                result = minScript_VarChangeType(@*params\index[i], #minScript_INTEGER)
                If result <> #minScript_OKAY
                  result = #minScript_INVALIDORMISSINGARRAYINDEX ;For consistency.
                  Break
                EndIf
              EndIf
              Select i-1
                Case 0
                  index0 = *params\index[i]\i
                Case 1
                  index1 = *params\index[i]\i
                Case 2
                  index2 = *params\index[i]\i
                Case 3
                  index3 = *params\index[i]\i
                Case 4
                  index4 = *params\index[i]\i
              EndSelect
            Next
            If result = #minScript_OKAY
              *ptrValue = minScript_VarGetArrayElement(*params\index[0], index0, index1, index2, index3, index4)
              If *ptrValue
                ;Now we attempt to copy the given value to the array.
                If *this\arrayType = #minScript_VARIANT
                  ;Free any existing value if appropriate.
                    If *ptrValue\variant\vt & #minScript_REFERENCE
                      If *ptrValue\variant\refPtr
                        result = *ptrValue\variant\refPtr\Clear()
                      EndIf
                    ElseIf *ptrValue\variant\vt = #minScript_STRING
                      minScript_StringFree(*ptrValue\variant\i)
                    EndIf
                    *ptrValue\variant\vt = *params\index[numParamsPassed-1]\vt
                    *ptrValue\variant\q = *params\index[numParamsPassed-1]\q
                    ;We simply zero the given variant (as opposed to clearing it) as we have done a direct copy.
                      *params\index[numParamsPassed-1]\vt = #minScript_EMPTY
                      *params\index[numParamsPassed-1]\q = 0
                Else
                  ;Attempt to promote the given value to the respective type.
                    result = minScript_VarChangeType(@*params\index[numParamsPassed-1], *this\vtConversionType)
                  If result = #minScript_OKAY
                    If *this\arrayType & #minScript_REFERENCE
                      ;Clear any existing object.
                      If *ptrValue\refPtr
                        result = *ptrValue\refPtr\Clear()
                      EndIf
                      *ptrValue\refPtr = *params\index[numParamsPassed-1]\refPtr
                    Else
                      Select *this\arrayType
                        Case #minScript_BYTE
                          *ptrValue\b = *params\index[numParamsPassed-1]\i
                        Case #minScript_WORD
                          *ptrValue\w = *params\index[numParamsPassed-1]\i
                        Case #minScript_CHARACTER
                          *ptrValue\c = *params\index[numParamsPassed-1]\i
                        Case #minScript_LONG
                          *ptrValue\l = *params\index[numParamsPassed-1]\i
                        Case #minScript_FLOAT
                          *ptrValue\f = *params\index[numParamsPassed-1]\r
                        Case #minScript_INTEGER
                          *ptrValue\i = *params\index[numParamsPassed-1]\i
                        Case #minScript_CURRENCY
                          *ptrValue\q = *params\index[numParamsPassed-1]\c
                        Case #minScript_REAL
                          *ptrValue\d = *params\index[numParamsPassed-1]\r
                        Case #minScript_STRING
                          minScript_StringFree(*ptrValue\i)
                          *ptrValue\i = *params\index[numParamsPassed-1]\i
                      EndSelect
                    EndIf
                    ;We simply zero the given variant (as opposed to clearing it) as we have done a direct copy.
                      *params\index[numParamsPassed-1]\vt = #minScript_EMPTY
                      *params\index[numParamsPassed-1]\q = 0
                  EndIf
                EndIf
              Else
                result = #minScript_INVALIDORMISSINGARRAYINDEX
              EndIf
            EndIf
          Else
            result = #minScript_INVALIDORMISSINGARRAYINDEX
          EndIf
        EndIf
      EndIf
    Case "redim"
      If invokeType = #minScript_METHOD_INVOKE
        If *this\blnBeingDestroyed = #False ;We don't return an error here.
          If numParamsPassed = 2
            If *params\index[1]\vt & #minScript_REFERENCE
              result = #minScript_INVALIDDATATYPE
            ElseIf *params\index[1]\vt <> #minScript_INTEGER
              result = minScript_VarChangeType(@*params\index[1], #minScript_INTEGER)
              If result <> #minScript_OKAY
              Else
                result = #minScript_INVALIDDATATYPE ;For consistency.
              EndIf
            EndIf
            If result = #minScript_OKAY
              If *params\index[1]\i >= 0
                ;We need to reset the relevant upper bound.
                  If *this\numDims
                    i = *this\numDims - 1
                  EndIf
                  *this\arrayUpperBounds[i] = *params\index[1]\i
                ;Now reset the array. We need to temporarily embed it within a minScript variant.
                  var\vt = #minScript_ARRAY
                  var\refPtr = *this
                result = minScript_VarSetArray(var, *this\arrayUpperBounds[0], *this\arrayUpperBounds[1], *this\arrayUpperBounds[2], *this\arrayUpperBounds[3], *this\arrayUpperBounds[4])
              Else
                result = #minScript_INVALIDPARAMETER
              EndIf
            EndIf
          Else
            result = #minScript_INCORRECTNUMBEROFPARAMETERS
          EndIf
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "ubound", "upperbound"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed > 2
          result = #minScript_TOOMANYPARAMETERS
        Else
          index0 = 1
          If numParamsPassed = 2
            ;Promote the given value to an integer type.
              result = minScript_VarChangeType(@*params\index[1], #minScript_INTEGER)
            If result = #minScript_OKAY
              index0 = *params\index[1]\i
            EndIf
          EndIf
          If result = #minScript_OKAY
            *ret\vt = #minScript_INTEGER
            If index0 >= 1 And index0 <=5
              *ret\i = *this\arrayUpperBounds[index0-1]
            Else
              *ret\i = -1
            EndIf
          EndIf
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Case "typeof"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          *ret\i = *this\arrayType
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "typeof2string"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          type$ = minScriptCodeClass_DISASSEMBLY_VT2String(*this\arrayType, codeObject)
          If type$ = ""
            type$ = "unknown"      
          EndIf
          result = minScript_VarSetString(*ret, type$, #False)
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "numelements"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          *ret\i = *this\numElements
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "numdims"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          *ret\i = *this\numDims
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "sizeofelement"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          *ret\i = *this\sizeOfElement
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Default
      result = #minScript_METHODORPROPERTYNOTSUPPORTED
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-PUBLIC ARRAY FUNCTIONS.
;These functions allow host apps to create their own minScript arrays for use with script variables etc.
;They are also used by the above class methods (and the minScript_RunTime_NewArray() function) to handle arrays.

;/////////////////////////////////////////////////////////////////////////////////
;The following initialises an array object. 
;We create the underlying minScriptObjectVariant object, though most of the class structure fields will be empty.
;Note that if *var points to an existing array with a refCount > 1, then the original array will remain in force
;and *var made to point to a new array.
;Note that we cannot reliably check the array type because it may be of a user-defined type (e.g. an array of objects).
;Any errors and the original variable is left untouched.
Procedure.i minScript_VarInitArray(*var.minScriptVariant, arrayType.u=#minScript_INTEGER)
  Protected result = #minScript_OKAY, *this.st_minScriptObjectVariantClass_Array, sizeOfElement, vtConversionType
  If *var
    ;Determine the size of individual elements.
      If arrayType & #minScript_REFERENCE
        sizeOfElement = SizeOf(INTEGER)
        vtConversionType = arrayType
      Else
        Select arrayType
          Case #minScript_BYTE
            sizeOfElement = SizeOf(BYTE)
            vtConversionType = #minScript_INTEGER
          Case #minScript_WORD
            sizeOfElement = SizeOf(WORD)
            vtConversionType = #minScript_INTEGER
          Case #minScript_CHARACTER
            sizeOfElement = SizeOf(CHARACTER)
            vtConversionType = #minScript_INTEGER
          Case #minScript_LONG
            sizeOfElement = SizeOf(LONG)
            vtConversionType = #minScript_INTEGER
          Case #minScript_FLOAT
            sizeOfElement = SizeOf(FLOAT)
            vtConversionType = #minScript_REAL
          Case #minScript_INTEGER
            sizeOfElement = SizeOf(INTEGER)
            vtConversionType = #minScript_INTEGER
          Case #minScript_CURRENCY
            sizeOfElement = SizeOf(QUAD)
            vtConversionType = #minScript_CURRENCY
          Case #minScript_REAL
            sizeOfElement = SizeOf(DOUBLE)
            vtConversionType = #minScript_REAL
          Case #minScript_STRING
            sizeOfElement = SizeOf(INTEGER)
            vtConversionType = #minScript_STRING
          Case #minScript_VARIANT
            sizeOfElement = SizeOf(minScriptVariant)
            vtConversionType = #minScript_VARIANT
          Default
            result = #minScript_UNRECOGNISEDVARIABLETYPE
        EndSelect          
      EndIf
    If result = #minScript_OKAY
      ;Allocate memory for a st_minScriptObjectVariantClass_Array structure.
        *this = AllocateStructure(st_minScriptObjectVariantClass_Array)
      If *this
        *this\vTable = ?VTable_minScriptObjectVariantClass_Array
        *this\arrayType = arrayType
        *this\sizeOfElement = sizeOfElement
        *this\vtConversionType = vtConversionType
        *this\refCount = 1
        ;Clear *var. We do not need to check for errors.
          minScript_VarClear(*var, #True)
        ;Set the *var fields.
          *var\vt = #minScript_ARRAY  ;Just in case!
          *var\refPtr = *this
      Else
        result = #minScript_OUTOFMEMORY      
      EndIf
    EndIf
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following sets an array's bounds and associated memory allocations. The array needs to have been initialized first through minScript_VarInitArray().
;Any existing array dimensions will be reset.
;If all upperBounds are negative then the array will be completely reset and the array data will be completely removed.
;Otherwise, any existing array data will be preserved as far as allowed by the new dimensions etc.
;This function allows for a max of 5 dimensions, though that limit is quite arbitrary.
Procedure.i minScript_VarSetArray(*var.minScriptVariant, upperBound0=-1, upperBound1=-1, upperBound2=-1, upperBound3=-1, upperBound4=-1)
  Protected result = #minScript_OKAY, *this.st_minScriptObjectVariantClass_Array, numDims, numElements, requiredMemSize, i
  Protected *ptr.st_minScript_ArrayData, t1
  If *var
    If *var\vt = #minScript_ARRAY And *var\refPtr
      *this = *var\refPtr
      ;Determine the number of dimensions etc.
        If upperBound0 >= 0
          numDims+1
          numElements = upperBound0 + 1
          If upperBound1 >= 0
            numDims+1
            numElements * (upperBound1 + 1)
            If upperBound2 >= 0
              numDims+1
              numElements * (upperBound2 + 1)
              If upperBound3 >= 0
                numDims+1
                numElements * (upperBound3 + 1)
                If upperBound4 >= 0
                  numDims+1
                  numElements * (upperBound4 + 1)
                Else
                  Goto minScript_VarSetArray_L3
                EndIf
              Else
                Goto minScript_VarSetArray_L2
              EndIf
            Else
              Goto minScript_VarSetArray_L1
            EndIf
          Else
            Goto minScript_VarSetArray_L0
          EndIf
        Else
          upperBound0 = -1
minScript_VarSetArray_L0:
          upperBound1 = -1
minScript_VarSetArray_L1:
          upperBound2 = -1
minScript_VarSetArray_L2:
          upperBound3 = -1
minScript_VarSetArray_L3:
          upperBound4 = -1
        EndIf
        ;Sort out any existing array data.
          requiredMemSize = numElements * *this\sizeOfElement
          If *this\pData
            ;First remove any redundant object or string variables.
            ;In such a case the memory requirements are being lowered and so there will be no problem reallocating memory.
            ;We can thus free surplus strings without worrying about reallocation.
              If *this\arrayType = #minScript_VARIANT
                For i = numElements To *this\numElements-1
                  If *this\pData\variant[i]\vt & #minScript_REFERENCE
                    If *this\pData\variant[i]\refPtr
                      t1 = *this\pData\variant[i]\refPtr\Clear()
                      If result = #minScript_OKAY
                        result = t1
                      EndIf
                    EndIf
                  ElseIf *this\pData\variant[i]\vt = #minScript_STRING
                    minScript_StringFree(*this\pData\variant[i]\sPtr)
                  EndIf
                Next              
              ElseIf *this\arrayType & #minScript_REFERENCE
                For i = numElements To *this\numElements-1
                  If *this\pData\refPtr[i] 
                    t1 = *this\pData\refPtr[i]\Clear()
                    If result = #minScript_OKAY
                      result = t1
                    EndIf
                  EndIf
                Next
              ElseIf *this\arrayType = #minScript_STRING
                For i = numElements To *this\numElements-1
                  minScript_StringFree(*this\pData\sPtr[i])
                Next
              EndIf
            ;Sort out memory allocations.
              If result = #minScript_OKAY And requiredMemSize
                ;Reallocate memory.
                  *ptr = ReAllocateMemory(*this\pData, requiredMemSize)
                  If *ptr
                    *this\pData = *ptr
                  Else ;Only possible if enlarging the size of the array's data section.
                    result = #minScript_OUTOFMEMORY
                  EndIf
              Else
                FreeMemory(*this\pData)
                *this\pData = 0
              EndIf
          ElseIf requiredMemSize ;Create new array data.
            *ptr = AllocateMemory(requiredMemSize)
            If *ptr
              *this\pData = *ptr
            Else
              result = #minScript_OUTOFMEMORY
            EndIf
          EndIf
        If result = #minScript_OKAY
          ;Fill in the fields of the st_keyScript_ReferenceObject structure.
            *this\numDims = numDims
            *this\numElements = numElements
            *this\arrayUpperBounds[0] = upperBound0
            *this\arrayUpperBounds[1] = upperBound1
            *this\arrayUpperBounds[2] = upperBound2
            *this\arrayUpperBounds[3] = upperBound3
            *this\arrayUpperBounds[4] = upperBound4
        EndIf
    Else
      result = #minScript_OBJECTNOTINITIALIZED
    EndIf
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns (subject to array bounds checks) the address of the given element in the specified array.
;Indexes given for non-existent dimensions are ignored.
;Returns zero if an error.
Procedure.i minScript_VarGetArrayElement(*var.minScriptVariant, index0, index1=-1, index2=-1, index3=-1, index4=-1)
  Protected address, *this.st_minScriptObjectVariantClass_Array, totOffset, i
  If *var And *var\vt = #minScript_ARRAY And *var\refPtr
    *this = *var\refPtr
    ;Validate indexes.
      If (index0 >=0 And index0 <= *this\arrayUpperBounds[0]) 
        If *this\numDims >= 2
          If (index1 < 0 Or index1 > *this\arrayUpperBounds[1])
            ProcedureReturn 0
          EndIf
        Else
          index1 = 0
        EndIf
        If *this\numDims >= 3
          If (index2 <0 Or index2 > *this\arrayUpperBounds[2])
            ProcedureReturn 0
          EndIf
        Else
          index2 = 0
        EndIf
        If *this\numDims >= 4
          If (index3 <0 Or index3 > *this\arrayUpperBounds[3])
            ProcedureReturn 0
          EndIf
        Else
          index3 = 0
        EndIf
        If *this\numDims = 5
          If (index4 <0 Or index4 > *this\arrayUpperBounds[4])
            ProcedureReturn 0
          EndIf
        Else
          index4 = 0
        EndIf
        ;Combine the indexes to give an offset for the element in question.
        ;totOffset = i0 + i1[(u0+1)] + i2[(u0+1)(u1+1)] + i3[(u0+1)(u1+1)(u2+1)] + i4[(u0+1)(u1+1)(u2+1)(u3+1)]
          totOffset = index4
          For i = 3 To 0 Step -1
            totOffset * (*this\arrayUpperBounds[i] + 1)
            Select i
              Case 3
                totOffset + index3
              Case 2
                totOffset + index2
              Case 1
                totOffset + index1
              Case 0
                totOffset + index0
            EndSelect           
          Next
        ;Now we can calculate the required address.
          address = *this\pData + totOffset * *this\sizeOfElement
      EndIf
  EndIf
  ProcedureReturn address
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


DataSection 
  VTable_minScriptObjectVariantClass_Array: 
    Data.i @minScriptObjectVariantClass_Array_Clear()
    Data.i @minScriptObjectVariantClass_Array_CopyToVariant()
    Data.i @minScriptObjectVariantClass_Array_Invoke()
EndDataSection



; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 120
; FirstLine = 116
; Folding = ---
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode