﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;HashMaps module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;     i)    minScript hash maps are user defined types. They are reference counted (not all user defined types need use reference counting).
;
;     ii)   A hashmap holds a series of 'properties' in the form of name/value pairs where each value can hold either a prescribed type of value or,
;           in the case of a variant type hashmap, any type of value.
;
;     iii)  Any attempt to retrieve a non-existent property will not result in a run-time error (just an empty return). 
;           However, any attempt to return the address of a non-existent property will result in an error.
;
;     iv)   Each hashmap has a 'current element' which can be accessed without refering to a property name (key). 
;           E.g. myHash() references the current element (if there is one). 
;
;NOTE to self. Variant hashmaps have \hashTypeVT set to #minScript_VARIANT, but individual elements do not have this bit set.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES

  Declare.i minScript_RunTime_NewHashmap(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariant, *ret.minScriptVariant)
  Declare minScriptHashmap_GCE(codeObject, uMsg)
  Declare minScriptObjectVariantClass_Hashmap_Clear(*this)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

  ;The following limit is arbitrary, but 10 should be enough!
  ;We do not use PB's PushMapPosition()/PopMapPosition() functions because they do not allow us to always determine the current position afterwards.
    #minScript_MAXHASHMAPPUSHES = 10
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES

  ;The following is the class template for the minScriptObjectVariant class representing our hashmaps.
    Structure st_minScriptObjectVariantClass_Hashmap
      *vTable
      refCount.i                          ;When this hits zero, the object is removed.
      hashTypeVT.i
      ;Property fields.
        Map properties.minScriptVariant()
        *currentElement.minScriptVariant  ;Allows scripts to query the 'current' element and iteration through the elements.
        ;We push the property names to best allow for deletions/insertions whilst enumerating all properties etc.
          pushedPropertyNames.s[#minScript_MAXHASHMAPPUSHES]
          numPushedProperties.i
      ;The following safeguard is probably not needed, but is intended to prevent the object being destroyed whilst it is being destroyed through a circular reference and some unforseen action that might increase the reference count etc.
        blnBeingDestroyed.i
    EndStructure

  ;The following holds info on individual hashmaps so that our GC extension can dispose of them if needs be.
    Structure st_minScriptHashmap_GCE_HashmapEntry
      codeObject.i
      *refPtr.st_minScriptObjectVariantClass_Hashmap
    EndStructure

  ;The following holds the information required to administer the class GC extension.
    Structure st_minScriptHashmap_GCE
      threadMutex.i
      Map hashmapObjects.st_minScriptHashmap_GCE_HashmapEntry()  
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS

  Global minScriptHashmapGCE_globals.st_minScriptHashmap_GCE
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following hashmap init function is called at startup to allow minScript to register the hashmap types etc.
;A host app would use similar code to register it's own user defined types.
Procedure.i minScript_InitHashmapTypes()
  Protected result = #minScript_OKAY
  ;Register the hashmap type.
    result = minScript_AddUserDefinedType("Hashmap", #minScript_HASHMAP)
  If result = #minScript_OKAY
    ;Register a run-time function used to create new instances of hashmap objects.
    ;This particular function is required to return an instance of a minScriptObjectVariant interface.
      result = minScript_AddRuntimeFunction("NEWHASHMAP", #minScript_STANDARDRUNTIME_NEWHASHMAP, @minScript_RunTime_NewHashmap())
    If result = #minScript_OKAY
      ;Create a mutex to protect access to the GCE data if appropriate.
        CompilerIf #PB_Compiler_Thread
          minScriptHashmapGCE_globals\threadMutex = CreateMutex()
          If minScriptHashmapGCE_globals\threadMutex = 0
            result = #minScript_OUTOFMEMORY
          EndIf
        CompilerEndIf
      If result = #minScript_OKAY
        ;Add appropriate GC extensions.
          minScript_AddGCExtension(@minScriptHashmap_GCE())
      EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our "NEWHASHMAP" runtime function used to initialise a hashmap object.
Procedure.i minScript_RunTime_NewHashmap(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *this.st_minScriptObjectVariantClass_Hashmap, hashTypeVT = #minScript_VARIANT, *ptrString.STRING
  If numParamsPassed = 1
    ;A hashtype has been specified.
    If *params\index[0]\vt = #minScript_STRING
      *ptrString = *params\index[0]
      hashTypeVT = minScriptCodeClass_HELPER_String2VT(codeObject, LCase(*ptrString\s))
    ElseIf *params\index[0]\vt = #minScript_INTEGER   
      hashTypeVT = *params\index[0]\i
      ;Check if the VT type is a registered type which we can allow.
      Select hashTypeVT
        Case #minScript_BYTE, #minScript_WORD, #minScript_CHARACTER, #minScript_LONG, #minScript_FLOAT
          hashTypeVT = 0
        Default
          If minScriptCodeClass_DISASSEMBLY_VT2String(hashTypeVT, codeObject) = ""
            hashTypeVT = 0
          EndIf
      EndSelect
    EndIf
  EndIf
  If hashTypeVT = 0
    result = #minScript_INVALIDDATATYPE
  ElseIf numParamsPassed = 0 Or numParamsPassed = 1
    ;Allocate memory for a st_minScriptObjectVariantClass_Hashmap structure.
      *this = AllocateStructure(st_minScriptObjectVariantClass_Hashmap)
    If *this
      *this\vTable = ?VTable_minScriptObjectVariantClass_Hashmap
      *this\refCount = 1
      *this\hashTypeVT = hashTypeVT
      ;Add to the hashmap GC extension.
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScriptHashmapGCE_globals\threadMutex)
      CompilerEndIf
        If AddMapElement(minScriptHashmapGCE_globals\hashmapObjects(), Str(*this))
          minScriptHashmapGCE_globals\hashmapObjects()\codeObject = codeObject
          minScriptHashmapGCE_globals\hashmapObjects()\refPtr = *this
          ;Set the *ret fields.
            *ret\vt = #minScript_HASHMAP
            *ret\refPtr = *this
        Else
          FreeStructure(*this)
          result = #minScript_OUTOFMEMORY
        EndIf
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScriptHashmapGCE_globals\threadMutex)
      CompilerEndIf
    Else
      result = #minScript_OUTOFMEMORY      
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS  
  EndIf
  ;We will clear #minScript_OUTOFMEMORY errors so that scripts can determine for themselves how to proceed. A script can always use the THROWERROR() run-time function.
  If result = #minScript_OUTOFMEMORY
    result = #minScript_OKAY
  EndIf  
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-GC Extension.

;/////////////////////////////////////////////////////////////////////////////////
;The following GC extension disposes of any hashmaps unaccounted for.
Procedure minScriptHashmap_GCE(codeObject, uMsg)
  Select uMsg
    Case #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED
      ;Not needed.
    Default ; #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET or #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED
      ;We simply loop through all outstanding hashmaps and free those belonging to the given code object.
        CompilerIf #PB_Compiler_Thread
          LockMutex(minScriptHashmapGCE_globals\threadMutex)
        CompilerEndIf
          ForEach minScriptHashmapGCE_globals\hashmapObjects()
            If minScriptHashmapGCE_globals\hashmapObjects()\codeObject = codeObject
              ;We forcibly delete the hashmap object by clearing all properties.
              ;We must avoid clearing any reference type properties which only leaves string types requiring clearing.
                ForEach minScriptHashmapGCE_globals\hashmapObjects()\refPtr\properties()
                  If minScriptHashmapGCE_globals\hashmapObjects()\refPtr\properties()\vt & #minScript_REFERENCE = 0
                    minScript_VarClear(minScriptHashmapGCE_globals\hashmapObjects()\refPtr\properties())
                  EndIf
                Next
              FreeStructure(minScriptHashmapGCE_globals\hashmapObjects()\refPtr)
              DeleteMapElement(minScriptHashmapGCE_globals\hashmapObjects())
            EndIf
          Next
        CompilerIf #PB_Compiler_Thread
          UnlockMutex(minScriptHashmapGCE_globals\threadMutex)
        CompilerEndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-minScriptObjectVariant CLASS METHODS.
;These 3 methods are required for scripts to be able to make use of our hash maps.

;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \Clear() method.
Procedure minScriptObjectVariantClass_Hashmap_Clear(*this.st_minScriptObjectVariantClass_Hashmap)
  Protected result = #minScript_OKAY, t1
  *this\refCount - 1
  If *this\refCount = 0 And *this\blnBeingDestroyed = #False
    *this\blnBeingDestroyed = #True
    ;Clear all properties.
      ForEach *this\properties()
        *this\currentElement = *this\properties() ;In case a property holds a class object and it's destructor accesses elements of the hashmap.
        t1 = minScript_VarClear(*this\properties())
        If result = #minScript_OKAY
          result = t1
        EndIf
      Next
    If *this\refCount = 0 ;In case some wayward class destructor increases the ref count of the array.
      FreeStructure(*this)
      ;Remove from the array GC extension.
        CompilerIf #PB_Compiler_Thread
          LockMutex(minScriptHashmapGCE_globals\threadMutex)
        CompilerEndIf
          DeleteMapElement(minScriptHashmapGCE_globals\hashmapObjects(), Str(*this))
        CompilerIf #PB_Compiler_Thread
          UnlockMutex(minScriptHashmapGCE_globals\threadMutex)
        CompilerEndIf
    Else
      *this\blnBeingDestroyed = #False    
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \CopyToVariant() method.
Procedure.i minScriptObjectVariantClass_Hashmap_CopyToVariant(*this.st_minScriptObjectVariantClass_Hashmap, *varDest.minScriptVariant)
  Protected result = #minScript_OKAY, i
  *varDest\vt = #minScript_HASHMAP ;Just in case!
  *varDest\refPtr = *this
  *this\refCount + 1
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptObjectVariantClass_Hashmap_Invoke(*this.st_minScriptObjectVariantClass_Hashmap, codeObject.minScriptCode, methodName$, invokeType, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, property$, *ptrValue.minScriptVariant, t1, type$
  Select methodName$
    Case "typeof"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          *ret\i = *this\hashTypeVT
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "typeof2string"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          type$ = minScriptCodeClass_DISASSEMBLY_VT2String(*this\hashTypeVT, codeObject)
          If type$ = ""
            type$ = "unknown"      
          EndIf
          result = minScript_VarSetString(*ret, type$, #False)
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "numelements", "numproperties"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          *ret\i = MapSize(*this\properties())
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "currentkey"
      If invokeType & #minScript_METHOD_GET
        If numParamsPassed = 1
            If *this\currentElement
              result = minScript_VarSetString(*ret, MapKey(*this\properties()), #False)
            EndIf
          Else
            result = #minScript_INCORRECTNUMBEROFPARAMETERS
          EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Case "clear"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 1
          If *this\blnBeingDestroyed = #False ;We don't return an error here.
            *this\blnBeingDestroyed = #True ;Will prevent any class destructors doing anything silly with these properties!
            ;Clear all properties.
              ForEach *this\properties()
                *this\currentElement = *this\properties() ;In case a property holds a class object and it's destructor accesses elements of the hashmap.
                t1 = minScript_VarClear(*this\properties())
                If result = #minScript_OKAY
                  result = t1
                EndIf
              Next
            *this\blnBeingDestroyed = #False
            ClearMap(*this\properties())
            *this\currentElement = 0
          EndIf        
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Case "nextelement", "nextproperty"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 1
          If *this\blnBeingDestroyed = #False ;We don't return an error here.
            *this\currentElement = NextMapElement(*this\properties())
            *ret\vt = #minScript_INTEGER
            *ret\i = *this\currentElement
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Case "popposition"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 1
          If *this\blnBeingDestroyed = #False ;We don't return an error here.
            If *this\numPushedProperties
              *this\numPushedProperties - 1
              ;Attempt to find the relevant property.
                If *this\pushedPropertyNames[*this\numPushedProperties]
                  If FindMapElement(*this\properties(), *this\pushedPropertyNames[*this\numPushedProperties])
                    *this\currentElement = *this\properties()
                  EndIf
                Else
                  *this\currentElement = 0
                  ResetMap(*this\properties())
                EndIf
            Else
              result = #minScript_HASHMAPPOSITIONNOTPUSHED
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Case "pushposition"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 1
          If *this\blnBeingDestroyed = #False ;We don't return an error here.
            If *this\numPushedProperties < #minScript_MAXHASHMAPPUSHES
              ;We need to retrieve the property name of the current element.
                If *this\currentElement
                  property$ = MapKey(*this\properties())
                EndIf
              ;Add to the pushedPropertyNames[] collection. 
                *this\pushedPropertyNames[*this\numPushedProperties] = property$
              *this\numPushedProperties + 1
            Else
              result = #minScript_TOOMANYHASHMAPPUSHES            
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Case "reset"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 1
          If *this\blnBeingDestroyed = #False ;We don't return an error here.
            *this\currentElement = 0
            ResetMap(*this\properties())
          EndIf        
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Case "deleteelement", "deleteproperty"
      If invokeType = #minScript_METHOD_INVOKE
        If *this\blnBeingDestroyed = #False ;We don't return an error here.
          ;2 valid options here. Either DeleteElement(*this, propertyName$) or DeleteElement(*this).
          If numParamsPassed = 1
            If *this\currentElement
              ;Clear the current element minScript variant.
                If *this\currentElement
                  result = minScript_VarClear(*this\properties())
                EndIf
              *this\currentElement = DeleteMapElement(*this\properties())   
              If result = #minScript_OKAY
                *ret\vt = #minScript_INTEGER
                *ret\i = *this\currentElement
              EndIf
            Else
              result = #minScript_NOCURRENTELEMENT
            EndIf
          ElseIf numParamsPassed = 2
            If *params\index[1]\vt = #minScript_STRING
              If *params\index[1]\sPtr
                property$ = PeekS(*params\index[1]\sPtr)
              EndIf
              If property$
                ;Good to go. 
                ;If the specified element exists then we have to clear the underlying minScript variant.
                If FindMapElement(*this\properties(), property$)
                  result = minScript_VarClear(*this\properties())
                  DeleteMapElement(*this\properties(), property$)
                  *this\currentElement = 0 ;After a successful named deletion, the map will have no current element.
                  ResetMap(*this\properties()) ;Just to make sure as I am unsure whether the same is true on OSX etc?
                  If result = #minScript_OKAY
                    *ret\vt = #minScript_INTEGER
                    *ret\i = 0
                  EndIf
                EndIf
              Else
                result = #minScript_SYNTAXERROR
              EndIf
            Else
              result = #minScript_INVALIDDATATYPE        
            EndIf
          Else
            result = #minScript_INCORRECTNUMBEROFPARAMETERS
          EndIf
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Case "findelement", "findproperty"
      If invokeType = #minScript_METHOD_INVOKE
        If *this\blnBeingDestroyed = #False ;We don't return an error here.
          If numParamsPassed = 2
            If *params\index[1]\vt = #minScript_STRING
              If *params\index[1]\sPtr
                property$ = PeekS(*params\index[1]\sPtr)
              EndIf
              If property$
                ;Good to go. 
                *ret\vt = #minScript_INTEGER
                  If FindMapElement(*this\properties(), property$)
                    *this\currentElement = *this\properties()
                    *ret\i = *this\currentElement
                  EndIf
              Else
                result = #minScript_SYNTAXERROR
              EndIf
            Else
              result = #minScript_INVALIDDATATYPE        
            EndIf
          Else
            result = #minScript_INCORRECTNUMBEROFPARAMETERS
          EndIf
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Default
      ;Now the properties defined on the fly!  
      If invokeType = #minScript_METHOD_SET
        If *this\blnBeingDestroyed = #False ;We don't return an error here.
          ;3 valid options here. Either set_property(*this, value) or set_indexedvalue(*this, value) or set_indexedvalue(*this, "property", value).
          ;Let  us retrieve the property name and a pointer to the value to be assigned.
          If methodName$ = #minScript_DEFAULTPROPERTY$
            ;Must have 2 or 3 parameters.
            If numParamsPassed = 2
              If *this\currentElement = 0
                result = #minScript_NOCURRENTELEMENT
              EndIf
              *ptrValue = *params\index[1]
            ElseIf numParamsPassed = 3
              If *params\index[1]\vt = #minScript_STRING
                If *params\index[1]\sPtr
                  property$ = PeekS(*params\index[1]\sPtr)
                EndIf
                If property$ = ""
                  result = #minScript_SYNTAXERROR
                EndIf
                *ptrValue = *params\index[2]
              Else
                result = #minScript_INVALIDDATATYPE        
              EndIf
            Else
              result = #minScript_INCORRECTNUMBEROFPARAMETERS
            EndIf
          Else ;set_property(*this, value)
            ;Must have 2 parameters.
            If numParamsPassed = 2
              property$ = methodName$    
              *ptrValue = *params\index[1]
            Else
              result = #minScript_INCORRECTNUMBEROFPARAMETERS
            EndIf
          EndIf
          If result = #minScript_OKAY
            If property$
              ;Good to go. First we find any existing property and clear it as appropriate. Otherwise we create a new property.
                *this\currentElement = FindMapElement(*this\properties(), property$)
            EndIf
            If *this\currentElement
              result = minScript_VarClear(*this\currentElement, #minScript_VARDONOTCLEARVTFIELD)
            Else
              *this\currentElement = AddMapElement(*this\properties(), property$)
              If *this\currentElement = 0
                result = #minscript_OUTOFMEMORY
              EndIf
            EndIf
            If result = #minScript_OKAY
              ;Do we need to promote the value.
                If *this\hashTypeVT <> #minScript_VARIANT
                  result = minScript_VarChangeType(*ptrValue, *this\hashTypeVT)
                EndIf
              If result = #minScript_OKAY
                ;We just copy the value across.
                *this\currentElement\vt = *ptrValue\vt
                *this\currentElement\q = *ptrValue\q
                ;Clear the given parameter. No need to free the variant as we have copied it.
                  *ptrValue\vt = #minScript_EMPTY
                  *ptrValue\q = 0
              EndIf
            EndIf
          EndIf
        EndIf
      Else ;METHOD_GET
        If invokeType&#minScript_METHOD_GET
          If Left(methodName$, 1) <> "@"
            ;3 valid options here. Either get_property(*this) or get_indexedvalue(*this) or get_indexedvalue(*this, "property").
            ;Let  us retrieve the property name and a pointer to the value to be assigned.
            If methodName$ = #minScript_DEFAULTPROPERTY$
              ;Must have 1 or 2 parameters.
              If numParamsPassed = 1
                If *this\currentElement = 0
                  result = #minScript_NOCURRENTELEMENT
                EndIf
              ElseIf numParamsPassed = 2
                If *params\index[1]\vt = #minScript_STRING
                  If *params\index[1]\sPtr
                    property$ = PeekS(*params\index[1]\sPtr)
                  EndIf
                  If property$ = ""
                    result = #minScript_SYNTAXERROR
                  EndIf
                Else
                  result = #minScript_INVALIDDATATYPE        
                EndIf
              Else
                result = #minScript_INCORRECTNUMBEROFPARAMETERS
              EndIf
            Else ;get_property(*this)
              ;Must have 1 parameters.
              If numParamsPassed = 1
                property$ = methodName$    
                If property$ = ""
                  result = #minScript_SYNTAXERROR
                EndIf
              Else
                result = #minScript_INCORRECTNUMBEROFPARAMETERS
              EndIf
            EndIf
            If result = #minScript_OKAY
              If *this\blnBeingDestroyed
                ;We save the map position in case this property is being retrieved through a class destructor whilst the hashmap is being deleted.
                PushMapPosition(*this\properties())
              EndIf
              If property$
                ;Good to go. Let us retrieve the specified property.
                  *ptrValue = FindMapElement(*this\properties(), property$)
              Else
                *ptrValue = *this\currentElement
              EndIf
              If *ptrValue
                ;We need to copy the property.
                result = minScript_VarCopy(*ret, *ptrValue, #minScript_VARDONOTFREEDESTINATIONVARIABLE)
              EndIf  
              If *this\blnBeingDestroyed
                ;Retrieve the map position in case this property is being retrieved through a class destructor whilst the hashmap is being deleted.
                  PopMapPosition(*this\properties())
              EndIf
              *this\currentElement = *this\properties()
            EndIf
          Else ;@GET
            methodName$ = Right(methodName$, Len(methodName$)-1)
            ;3 valid options here. Either @get_property(*this) or @get_indexedvalue(*this, "property") or @get_indexedvalue(*this)
            ;Let  us retrieve the property name and a pointer to the value to be assigned.
            If methodName$ = #minScript_DEFAULTPROPERTY$
              ;Must have 1 or 2 parameters.
              If numParamsPassed = 1
                If *this\currentElement = 0
                  result = #minScript_NOCURRENTELEMENT
                EndIf
              ElseIf numParamsPassed = 2
                If *params\index[1]\vt = #minScript_STRING
                  If *params\index[1]\sPtr
                    property$ = PeekS(*params\index[1]\sPtr)
                  EndIf
                  If property$ = ""
                    result = #minScript_SYNTAXERROR
                  EndIf
                Else
                  result = #minScript_INVALIDDATATYPE        
                EndIf
              Else
                result = #minScript_INCORRECTNUMBEROFPARAMETERS
              EndIf
            Else ;@get_property(*this)
              ;Must have 1 parameters.
              If numParamsPassed = 1
                property$ = methodName$    
                If property$ = ""
                  result = #minScript_SYNTAXERROR
                EndIf
              Else
                result = #minScript_INCORRECTNUMBEROFPARAMETERS
              EndIf
            EndIf
            If result = #minScript_OKAY
              If *this\blnBeingDestroyed
                ;We save the map position in case this property is being retrieved through a class destructor whilst the hashmap is being deleted.
                PushMapPosition(*this\properties())
              EndIf
              If property$
                ;Good to go. Let us retrieve the specified property.
                  *ptrValue = FindMapElement(*this\properties(), property$)
                  If *ptrValue = 0
                    result = #minScript_METHODORPROPERTYNOTSUPPORTED
                  EndIf
              Else
                *ptrValue = *this\currentElement
              EndIf
              If *ptrValue
                *ret\vt = #minScript_INTEGER
                If *ptrValue\vt = #minScript_STRING
                  *ret\i = *ptrValue\sPtr
                Else
                  *ret\i = *ptrValue
                EndIf
              EndIf  
              If *this\blnBeingDestroyed
                ;Retrieve the map position in case this property is being retrieved through a class destructor whilst the hashmap is being deleted.
                  PopMapPosition(*this\properties())
              EndIf
              *this\currentElement = *this\properties()
            EndIf
          EndIf
        Else
          result = #minScript_METHODORPROPERTYNOTSUPPORTED
        EndIf
      EndIf
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



DataSection 
  VTable_minScriptObjectVariantClass_Hashmap: 
    Data.i @minScriptObjectVariantClass_Hashmap_Clear()
    Data.i @minScriptObjectVariantClass_Hashmap_CopyToVariant()
    Data.i @minScriptObjectVariantClass_Hashmap_Invoke()
EndDataSection



; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 97
; FirstLine = 93
; Folding = --
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode