﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Pointers module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;     i)    This module implements basic pointers at the user-defined refence object level and so will run a little faster than pointers implemented in script.
;           Each pointer object can be dereferenced to set/retrieve byte, word, character, long, float, integer, currency and real (double) values.
;
;     ii)   As with the majority of our user-defined types, these pointers are reference counted which might strike some as a bit odd, but we had no choice.
;           If we made physical copies of a pointer each time a 'copy' action was performed by minScript then there would be no way for a script to set the
;           properties of the original pointer because we would always be operating on separate copies of the original pointer.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES

  Declare.i minScript_RunTime_NewPointer(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariant, *ret.minScriptVariant)
  Declare minScriptPointer_GCE(codeObject, uMsg)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES

  ;The following is the class template for the minScriptObjectVariant class representing our pointers.
    Structure st_minScriptObjectVariantClass_Pointer
      *vTable
      refCount.i
      codeObject.i
      startSetWithOffsets$    ;To help out with a set with offsets. A minScript variant type.
      *address.st_minScript_memoryPointer
    EndStructure

  ;The following holds info on individual pointers so that our GC extension can dispose of them if needs be.
    Structure st_minScriptPointer_GCE_Pointers
      codeObject.i
      *refPtr.st_minScriptObjectVariantClass_Pointer
    EndStructure

  ;The following holds the information required to administer the class GC extension.
    Structure st_minScriptPointer_GCE
      threadMutex.i
      List pointerObjects.st_minScriptPointer_GCE_Pointers()  
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS

  Global minScriptPointerGCE_globals.st_minScriptPointer_GCE
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following pointer init function is called at startup to allow minScript to register the pointer types etc.
;A host app would use similar code to register it's own user defined types.
Procedure.i minScript_InitPointerTypes()
  Protected result = #minScript_OKAY
  ;Register the pointer type.
    result = minScript_AddUserDefinedType("Pointer", #minScript_POINTER)
  If result = #minScript_OKAY
    ;Register a run-time function used to create new instances of pointer objects.
    ;This particular function is required to return an instance of a minScriptObjectVariant interface.
      result = minScript_AddRuntimeFunction("NEWPOINTER", #minScript_STANDARDRUNTIME_NEWPOINTER, @minScript_RunTime_NewPointer())
    If result = #minScript_OKAY
      ;Create a mutex to protect access to the GCE data if appropriate.
        CompilerIf #PB_Compiler_Thread
          minScriptPointerGCE_globals\threadMutex = CreateMutex()
          If minScriptPointerGCE_globals\threadMutex = 0
            result = #minScript_OUTOFMEMORY
          EndIf
        CompilerEndIf
      If result = #minScript_OKAY
        ;Add appropriate GC extensions.
          minScript_AddGCExtension(@minScriptPointer_GCE())
      EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our "NEWPOINTER" runtime function used to initialise a pointer object.
Procedure.i minScript_RunTime_NewPointer(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *this.st_minScriptObjectVariantClass_Pointer, address
  If numParamsPassed <= 1
    If numParamsPassed = 1
      If *params\index[0]\vt = #minScript_INTEGER   
        address = *params\index[0]\i
      Else
        result = #minScript_INVALIDDATATYPE
      EndIf
    EndIf
    If result = #minScript_OKAY
      ;Allocate memory for a st_minScriptObjectVariantClass_Pointer structure.
        *this = AllocateStructure(st_minScriptObjectVariantClass_Pointer)
      If *this
        *this\vTable = ?VTable_minScriptObjectVariantClass_Pointer
        *this\refCount = 1
        *this\codeObject = codeObject
        *this\address = address
        ;Add to the pointer GC extension.
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScriptPointerGCE_globals\threadMutex)
          CompilerEndIf
            If AddElement(minScriptPointerGCE_globals\pointerObjects())
              minScriptPointerGCE_globals\pointerObjects()\codeObject = codeObject
              minScriptPointerGCE_globals\pointerObjects()\refPtr = *this
              *ret\vt = #minScript_POINTER
              *ret\refPtr = *this
            Else ;Should free the memory just in case.
              FreeStructure(*this)
              result = #minScript_OUTOFMEMORY
            EndIf
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScriptPointerGCE_globals\threadMutex)
          CompilerEndIf
      Else
        result = #minScript_OUTOFMEMORY
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS  
  EndIf
  ;We will clear #minScript_OUTOFMEMORY errors so that scripts can determine for themselves how to proceed. A script can always use the THROWERROR() run-time function.
  If result = #minScript_OUTOFMEMORY
    result = #minScript_OKAY
  EndIf  
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-GC Extension.

;/////////////////////////////////////////////////////////////////////////////////
;The following GC extension disposes of any pointers unaccounted for.
Procedure minScriptPointer_GCE(codeObject, uMsg)
  Select uMsg
    Case #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED
      ;Not needed.
    Default ; #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET or #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED
      ;We simply loop through all outstanding pointers and free those belonging to the given code object.
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScriptPointerGCE_globals\threadMutex)
      CompilerEndIf
        ForEach minScriptPointerGCE_globals\pointerObjects()
          If minScriptPointerGCE_globals\pointerObjects()\codeObject = codeObject
            FreeStructure(minScriptPointerGCE_globals\pointerObjects()\refPtr)
            DeleteElement(minScriptPointerGCE_globals\pointerObjects())
          EndIf           
        Next
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScriptPointerGCE_globals\threadMutex)
      CompilerEndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-minScriptObjectVariant CLASS METHODS.
;These 3 methods are required for scripts to be able to make use of our pointers.

;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \Clear() method.
Procedure minScriptObjectVariantClass_Pointer_Clear(*this.st_minScriptObjectVariantClass_Pointer)
  Protected result = #minScript_OKAY
  *this\refCount - 1
  If *this\refCount = 0 
    ;Remove from the GCE memory list.
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScriptPointerGCE_globals\threadMutex)
      CompilerEndIf
        ForEach minScriptPointerGCE_globals\pointerObjects()
          If minScriptPointerGCE_globals\pointerObjects()\refPtr = *this
            DeleteElement(minScriptPointerGCE_globals\pointerObjects())
            Break
          EndIf           
        Next
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScriptPointerGCE_globals\threadMutex)
      CompilerEndIf
    FreeStructure(*this)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \CopyToVariant() method.
Procedure.i minScriptObjectVariantClass_Pointer_CopyToVariant(*this.st_minScriptObjectVariantClass_Pointer, *varDest.minScriptVariant)
  Protected result = #minScript_OKAY
  *varDest\vt = #minScript_POINTER ;Just in case!
  *varDest\refPtr = *this
  *this\refCount + 1
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptObjectVariantClass_Pointer_Invoke(*this.st_minScriptObjectVariantClass_Pointer, codeObject.minScriptCode, methodName$, invokeType, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrValue.st_minScript_memoryPointer, offset
  If *this\startSetWithOffsets$
    If methodName$ = #minScript_DEFAULTPROPERTY$ And invokeType = #minScript_METHOD_SET
      If numParamsPassed = 3
        If *this\address
          If *params\index[1]\vt & #minScript_REFERENCE Or *params\index[2]\vt & #minScript_REFERENCE
            result = #minScript_INVALIDDATATYPE
          Else
            minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
            Select *this\startSetWithOffsets$
              Case "i"
                minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(INTEGER)
                *ptrValue\i = *params\index[2]\i
              Case "b"
                minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(BYTE)
                *ptrValue\b = *params\index[2]\i
              Case "w"
                minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(WORD)
                *ptrValue\w = *params\index[2]\i
              Case "c"
                minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(CHARACTER)
                *ptrValue\c = *params\index[2]\i
              Case "l"
                minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(LONG)
                *ptrValue\l = *params\index[2]\i
              Case "f"
                minScript_VarChangeType(*params\index[2], #minScript_REAL)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(FLOAT)
                *ptrValue\f = *params\index[2]\r
              Case "cur"
                minScript_VarChangeType(*params\index[2], #minScript_CURRENCY)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(QUAD)
                *ptrValue\q = *params\index[2]\q
              Case "r"
                minScript_VarChangeType(*params\index[2], #minScript_REAL)
                *ptrValue = *this\address + *params\index[1]\i * SizeOf(DOUBLE)
                *ptrValue\d = *params\index[2]\r
            EndSelect
          EndIf
        Else
          result = #minScript_NULLPOINTER
        EndIf
      Else
        result = #minScript_INCORRECTNUMBEROFPARAMETERS
      EndIf
    Else
      result = #minScript_METHODORPROPERTYNOTSUPPORTED
    EndIf
    *this\startSetWithOffsets$ = ""
  Else
    Select methodName$
      Case "address"
        If invokeType & #minScript_METHOD_GET
          If numParamsPassed = 1
            *ret\vt = #minScript_INTEGER
            *ret\i = *this\address
          Else
            result = #minScript_INCORRECTNUMBEROFPARAMETERS
          EndIf
        ElseIf invokeType = #minScript_METHOD_SET
          If numParamsPassed = 2
            If *params\index[1]\vt & #minScript_REFERENCE
              result = #minScript_INVALIDDATATYPE
            Else
              minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
              *this\address = *params\index[1]\i
            EndIf
          Else
            result = #minScript_INCORRECTNUMBEROFPARAMETERS
          EndIf
        EndIf
      Case "b", "w", "c", "l", "i", "f", "cur", "r"
        Select invokeType
          Case #minScript_METHOD_GET
            If numParamsPassed = 1
              If *this\address
                Select methodName$
                  Case "b"
                    *ret\vt = #minScript_INTEGER
                    *ret\i = *this\address\b
                  Case "w"
                    *ret\vt = #minScript_INTEGER
                    *ret\i = *this\address\w
                  Case "c"
                    *ret\vt = #minScript_INTEGER
                    *ret\i = *this\address\c
                  Case "l"
                    *ret\vt = #minScript_INTEGER
                    *ret\i = *this\address\l
                  Case "i"
                    *ret\vt = #minScript_INTEGER
                    *ret\i = *this\address\i
                  Case "f"
                    *ret\vt = #minScript_REAL
                    *ret\r = *this\address\f
                  Case "cur"
                    *ret\vt = #minScript_CURRENCY
                    *ret\q = *this\address\q
                  Case "r"
                    *ret\vt = #minScript_REAL
                    *ret\r = *this\address\d
                EndSelect
              Else
                result = #minScript_NULLPOINTER
              EndIf
            Else
              result = #minScript_INCORRECTNUMBEROFPARAMETERS
            EndIf
          Case #minScript_METHOD_INVOKE ;This is effectively a GET with an offset parameter.
            If numParamsPassed <= 2
              If numParamsPassed = 2 ;An offset has been specified.
                If *params\index[1]\vt & #minScript_REFERENCE
                  result = #minScript_INVALIDDATATYPE
                Else
                  minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
                  offset = *params\index[1]\i
                EndIf
              EndIf
              If result = #minScript_OKAY
                If *this\address
                  *ret\vt = #minScript_INTEGER
                  *ptrValue = *this\address
                  Select methodName$
                    Case "b"
                      *ptrValue + offset * SizeOf(BYTE)
                      *ret\vt = #minScript_INTEGER
                      *ret\i = *ptrValue\b
                    Case "w"
                      *ptrValue + offset * SizeOf(WORD)
                      *ret\vt = #minScript_INTEGER
                      *ret\i = *ptrValue\w
                    Case "c"
                      *ptrValue + offset * SizeOf(CHARACTER)
                      *ret\vt = #minScript_INTEGER
                      *ret\i = *ptrValue\c
                    Case "l"
                      *ptrValue + offset * SizeOf(LONG)
                      *ret\vt = #minScript_INTEGER
                      *ret\i = *ptrValue\l
                    Case "i"
                      *ptrValue + offset * SizeOf(INTEGER)
                      *ret\vt = #minScript_INTEGER
                      *ret\i = *ptrValue\i
                    Case "f"
                      *ptrValue + offset * SizeOf(FLOAT)
                      *ret\vt = #minScript_REAL
                      *ret\r = *ptrValue\f
                    Case "cur"
                      *ptrValue + offset * SizeOf(QUAD)
                      *ret\vt = #minScript_CURRENCY
                      *ret\q = *ptrValue\q
                    Case "r"
                      *ptrValue + offset * SizeOf(DOUBLE)
                      *ret\vt = #minScript_REAL
                      *ret\r = *ptrValue\d
                   EndSelect
                Else
                  result = #minScript_NULLPOINTER
                EndIf
              EndIf
            Else
              result = #minScript_INCORRECTNUMBEROFPARAMETERS
            EndIf
          Case #minScript_METHOD_SET ;No offsets.
            If numParamsPassed = 2
              If *params\index[1]\vt & #minScript_REFERENCE
                result = #minScript_INVALIDDATATYPE
              ElseIf *this\address
                Select methodName$
                  Case "b"
                    minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
                    *this\address\b = *params\index[1]\i
                  Case "w"
                    minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
                    *this\address\w = *params\index[1]\i
                  Case "c"
                    minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
                    *this\address\c = *params\index[1]\i
                  Case "l"
                    minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
                    *this\address\l = *params\index[1]\i
                  Case "i"
                    minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
                    *this\address\i = *params\index[1]\i
                  Case "f"
                    minScript_VarChangeType(*params\index[1], #minScript_REAL)
                    *this\address\f = *params\index[1]\r
                  Case "cur"
                    minScript_VarChangeType(*params\index[1], #minScript_CURRENCY)
                    *this\address\q = *params\index[1]\q
                  Case "r"
                    minScript_VarChangeType(*params\index[1], #minScript_REAL)
                    *this\address\d = *params\index[1]\r
                EndSelect
              Else
                result = #minScript_NULLPOINTER
              EndIf
            Else
              result = #minScript_INCORRECTNUMBEROFPARAMETERS
            EndIf
          Case #minScript_METHOD_GETOBJECT 
            ;The beginning of a set with offsets which requires us to return the value of *this as an object.
            ;This will then be followed up with a set default value etc. This is just the way our parser has to handle this because of support for default properties.
            *this\refCount + 1
            *this\startSetWithOffsets$ = methodName$
            *ret\VT  = #minScript_POINTER
            *ret\refPtr = *this
        EndSelect
      Default
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
    EndSelect
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



DataSection 
  VTable_minScriptObjectVariantClass_Pointer: 
    Data.i @minScriptObjectVariantClass_Pointer_Clear()
    Data.i @minScriptObjectVariantClass_Pointer_CopyToVariant()
    Data.i @minScriptObjectVariantClass_Pointer_Invoke()
EndDataSection



; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 81
; FirstLine = 74
; Folding = --
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode