﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;StringBuilder module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;     i)    minScript string builders are user defined types. They are reference counted (not all user defined types need use reference counting).
;
;     ii)   We use our memory file utility class as the basis of each string builder and start the memory file with 4 bytes which will hold the byte length of the string.
;           This keeps the string builder internal format identical to a minScript string and can make serialising a string builder a simple matter
;           of copying the address etc. 
;           We use a memory file for convenience since minScript makes use of it anyhow.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES

  Declare.i minScript_RunTime_NewStringBuilder(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariant, *ret.minScriptVariant)
  Declare minScriptSTringBuilder_GCE(codeObject, uMsg)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES

  ;The following is the class template for the minScriptObjectVariant class representing our StringBuilders.
    Structure st_minScriptObjectVariantClass_StringBuilder
      *vTable
      refCount.i
      codeObject.i
      ;Memory file fields.
        memoryFile.MemoryFile
        initialSize.i
        pageSize.i
    EndStructure

  ;The following holds info on individual string builders so that our GC extension can dispose of them if needs be.
    Structure st_minScriptStringBuilder_GCE_StringBuilderEntry
      codeObject.i
      *refPtr.st_minScriptObjectVariantClass_StringBuilder
    EndStructure

  ;The following holds the information required to administer the GC extension.
    Structure st_minScriptStringBuilder_GCE
      threadMutex.i
      List stringBuilderObjects.st_minScriptStringBuilder_GCE_StringBuilderEntry()  
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS

  Global minScriptStringBuilderGCE_globals.st_minScriptStringBuilder_GCE
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following string builder init function is called at startup to allow minScript to register the string builder types etc.
;A host app would use similar code to register it's own user defined types.
Procedure.i minScript_InitStringBuilderTypes()
  Protected result = #minScript_OKAY
  ;Register the string builder type.
    result = minScript_AddUserDefinedType("StringBuilder", #minScript_STRINGBUILDER)
  If result = #minScript_OKAY
    ;Register a run-time function used to create new instances of string builder objects.
    ;This particular function is required to return an instance of a minScriptObjectVariant interface.
      result = minScript_AddRuntimeFunction("NEWSTRINGBUILDER", #minScript_STANDARDRUNTIME_NEWSTRINGBUILDER, @minScript_RunTime_NewStringBuilder())
    If result = #minScript_OKAY
      ;Create a mutex to protect access to the GCE data if appropriate.
        CompilerIf #PB_Compiler_Thread
          minScriptStringBuilderGCE_globals\threadMutex = CreateMutex()
          If minScriptStringBuilderGCE_globals\threadMutex = 0
            result = #minScript_OUTOFMEMORY
          EndIf
        CompilerEndIf
      If result = #minScript_OKAY
        ;Add appropriate GC extensions.
          minScript_AddGCExtension(@minScriptStringBuilder_GCE())
      EndIf
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our "NEWSTRINGBUILDER" runtime function used to initialise a string builder object.
Procedure.i minScript_RunTime_NewStringBuilder(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *this.st_minScriptObjectVariantClass_StringBuilder, initialSize = 1024
  If numParamsPassed <= 1
    If numParamsPassed = 1
      If *params\index[0]\vt = #minScript_INTEGER   
        If *params\index[0]\i > 0
          initialSize = *params\index[0]\i
        EndIf
      Else
        result = #minScript_INVALIDDATATYPE
      EndIf
    EndIf
    If result = #minScript_OKAY
      ;Allocate memory for a st_minScriptObjectVariantClass_StringBuilder structure.
        *this = AllocateStructure(st_minScriptObjectVariantClass_StringBuilder)
      If *this
        *this\vTable = ?VTable_minScriptObjectVariantClass_StringBuilder
        *this\refCount = 1
        *this\codeObject = codeObject
        *this\memoryFile = NewMemoryFile(initialSize+SizeOf(LONG), 1024, #MEMORYFILE_VARIABLEBUFFER)
        If *this\memoryFile
          ;Write an empty 4-bytes into which we may later write the total byte length to keep compatibility with minScript strings.
          ;No need to check for errors.
            *this\memoryFile\WriteLong(0)
          *this\pageSize = 1024
          *this\initialSize = initialSize
          ;Add to the string builder GC extension.
            CompilerIf #PB_Compiler_Thread
              LockMutex(minScriptStringBuilderGCE_globals\threadMutex)
            CompilerEndIf
              If AddElement(minScriptStringBuilderGCE_globals\stringBuilderObjects())
                minScriptStringBuilderGCE_globals\stringBuilderObjects()\codeObject = codeObject
                minScriptStringBuilderGCE_globals\stringBuilderObjects()\refPtr = *this
                *ret\vt = #minScript_STRINGBUILDER
                *ret\refPtr = *this
              Else ;Should free the memory just in case.
                *this\memoryFile\Destroy()
                Goto minScript_RunTime_NewStringBuilder_L1
              EndIf
            CompilerIf #PB_Compiler_Thread
              UnlockMutex(minScriptStringBuilderGCE_globals\threadMutex)
            CompilerEndIf
        Else
minScript_RunTime_NewStringBuilder_L1:
          FreeStructure(*this)
          result = #minScript_OUTOFMEMORY
        EndIf
      Else
        result = #minScript_OUTOFMEMORY
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS  
  EndIf
  ;We will clear #minScript_OUTOFMEMORY errors so that scripts can determine for themselves how to proceed. A script can always use the THROWERROR() run-time function.
  If result = #minScript_OUTOFMEMORY
    result = #minScript_OKAY
  EndIf  
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-GC Extension.

;/////////////////////////////////////////////////////////////////////////////////
;The following GC extension disposes of any string builders unaccounted for.
Procedure minScriptStringBuilder_GCE(codeObject, uMsg)
  Select uMsg
    Case #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED
      ;Not needed.
    Default ; #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET or #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED
      ;We simply loop through all outstanding string builders and free those belonging to the given code object.
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScriptStringBuilderGCE_globals\threadMutex)
      CompilerEndIf
        ForEach minScriptStringBuilderGCE_globals\stringBuilderObjects()
          If minScriptStringBuilderGCE_globals\stringBuilderObjects()\codeObject = codeObject
            If minScriptStringBuilderGCE_globals\stringBuilderObjects()\refPtr\memoryFile
              minScriptStringBuilderGCE_globals\stringBuilderObjects()\refPtr\memoryFile\Destroy()
            EndIf
            FreeStructure(minScriptStringBuilderGCE_globals\stringBuilderObjects()\refPtr)
            DeleteElement(minScriptStringBuilderGCE_globals\stringBuilderObjects())
          EndIf           
        Next
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScriptStringBuilderGCE_globals\threadMutex)
      CompilerEndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-minScriptObjectVariant CLASS METHODS.
;These 3 methods are required for scripts to be able to make use of our string builders.

;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \Clear() method.
Procedure minScriptObjectVariantClass_StringBuilder_Clear(*this.st_minScriptObjectVariantClass_StringBuilder)
  Protected result = #minScript_OKAY
  *this\refCount - 1
  If *this\refCount = 0 
    ;Remove from the GCE memory list.
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScriptStringBuilderGCE_globals\threadMutex)
      CompilerEndIf
        ForEach minScriptStringBuilderGCE_globals\stringBuilderObjects()
          If minScriptStringBuilderGCE_globals\stringBuilderObjects()\refPtr = *this
            DeleteElement(minScriptStringBuilderGCE_globals\stringBuilderObjects())
            Break
          EndIf           
        Next
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScriptStringBuilderGCE_globals\threadMutex)
      CompilerEndIf
    If *this\memoryFile
      *this\memoryFile\Destroy()
    EndIf
    FreeStructure(*this)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \CopyToVariant() method.
Procedure.i minScriptObjectVariantClass_StringBuilder_CopyToVariant(*this.st_minScriptObjectVariantClass_StringBuilder, *varDest.minScriptVariant)
  Protected result = #minScript_OKAY
  *varDest\vt = #minScript_STRINGBUILDER ;Just in case!
  *varDest\refPtr = *this
  *this\refCount + 1
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptObjectVariantClass_StringBuilder_Invoke(*this.st_minScriptObjectVariantClass_StringBuilder, codeObject.minScriptCode, methodName$, invokeType, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, pageSize = 1024, *ptrValue.st_minScript_memoryPointer, numBytes
  Select methodName$
    Case "clear"
      If invokeType = #minScript_METHOD_INVOKE
        *this\memoryFile\Reset()
        *this\memoryFile\WriteLong(0)
        *this\pageSize = 1024
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "pagesize"
      If invokeType = #minScript_METHOD_GET  
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          If *this\memoryFile
            *ret\i = *this\memoryFile\GetPageSize()
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_SET
        If numParamsPassed = 2
          If *params\index[1]\vt & #minScript_REFERENCE
            result = #minScript_INVALIDDATATYPE
          Else
            If *this\memoryFile
              minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
              If *params\index[1]\i > 0
                pageSize = *params\index[1]\i
              EndIf
              *this\memoryFile\SetPageSize(pageSize)
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "appendcharacterbuffer"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 2
          *ret\vt = #minScript_INTEGER
          If *this\memoryFile
            If *params\index[1]\vt = #minScript_INTEGER
              ;We need to identify the number of bytes in the character buffer.
              If *params\index[1]\i
                *ret\i = 1 ;Signals ok.
                *ptrValue = *params\index[1]\i
                While *ptrValue\c
                  *ptrValue + SizeOf(CHARACTER)
                Wend
                numBytes = *ptrValue - *params\index[1]\i
                If numBytes
                  *this\memoryFile\FileSeek(*this\memoryFile\Lof())
                  *ret\i = *this\memoryFile\WriteData(*params\index[1]\i, numBytes)
                EndIf
              Else
                result = #minScript_NULLPOINTER
              EndIf
            Else
              result = #minScript_INVALIDDATATYPE
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "appendstring"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 2
          *ret\vt = #minScript_INTEGER
          If *this\memoryFile
            result = minScript_VarChangeType(*params\index[1], #minScript_STRING)
            If result = #minScript_OKAY
              *ret\i = 1 ;Signals ok.
              If *params\index[1]\sPtr
                *ptrValue = *params\index[1]\sPtr - SizeOf(LONG)
                *this\memoryFile\FileSeek(*this\memoryFile\Lof())
                *ret\i = *this\memoryFile\WriteData(*params\index[1]\sPtr, *ptrValue\l)
              EndIf
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "appendstringbyaddress"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 2
          *ret\vt = #minScript_INTEGER
          If *this\memoryFile
            If *params\index[1]\vt = #minScript_INTEGER
              If *params\index[1]\i
                *ret\i = 1 ;Signals ok.
                *ptrValue = *params\index[1]\i - SizeOf(LONG)
                *this\memoryFile\FileSeek(*this\memoryFile\Lof())
                *ret\i = *this\memoryFile\WriteData(*params\index[1]\i, *ptrValue\l)
              Else
                result = #minScript_NULLPOINTER
              EndIf
            Else
              result = #minScript_INVALIDDATATYPE
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "tostring"
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 1
          *ret\vt = #minScript_STRING
          If *this\memoryFile
            numBytes = *this\memoryFile\Lof() - SizeOf(LONG)
            If numBytes
              ;We create the string semi-manually to allow for the addition of a null-terminator.
              *ret\sPtr = minScript_StringAllocLen(numBytes >> (SizeOf(CHARACTER)-1))
              If *ret\sPtr
                CopyMemory(*this\memoryFile\GetFileMemoryBase() + SizeOf(LONG), *ret\sPtr, numBytes)
              Else
                result = #minScript_OUTOFMEMORY
              EndIf
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "fasttostring"
      ;Here we simply point the return string to the memory file (after a single modification). We subsequently have to invalidate the underlying memory file.
      If invokeType = #minScript_METHOD_INVOKE
        If numParamsPassed = 1
          *ret\vt = #minScript_STRING
          If *this\memoryFile
            ;Need to ensure the memory is null terminated. 
            *this\memoryFile\FileSeek(*this\memoryFile\Lof())
            If *this\memoryFile\WriteCharacter(0)
              *ptrValue = *this\memoryFile\GetFileMemoryBase()
              *ptrValue\l = *this\memoryFile\Lof() - SizeOf(LONG) - SizeOf(CHARACTER)
              *ret\sPtr = *ptrValue + SizeOf(LONG)
              *this\memoryFile\Destroy(#False)
              *this\memoryFile = 0
            Else
              result = #minScript_OUTOFMEMORY
            EndIf
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED 
      EndIf
    Case "length"
      If invokeType = #minScript_METHOD_GET  
        If numParamsPassed = 1
          *ret\vt = #minScript_INTEGER
          If *this\memoryFile
            *ret\i = (*this\memoryFile\Lof() - SizeOf(LONG)) >> (SizeOf(CHARACTER)-1)
          EndIf
        Else
          result = #minScript_INCORRECTNUMBEROFPARAMETERS
        EndIf
      ElseIf invokeType = #minScript_METHOD_INVOKE
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      Else
        result = #minScript_PROPERTYISREADONLY
      EndIf
    Default
      result = #minScript_METHODORPROPERTYNOTSUPPORTED
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



DataSection 
  VTable_minScriptObjectVariantClass_StringBuilder: 
    Data.i @minScriptObjectVariantClass_StringBuilder_Clear()
    Data.i @minScriptObjectVariantClass_StringBuilder_CopyToVariant()
    Data.i @minScriptObjectVariantClass_StringBuilder_Invoke()
EndDataSection



; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 83
; FirstLine = 76
; Folding = --
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode