﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Variants Module.
;
;© Big Eared Software 2016.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;minScript strings are stored in a synthesised BSTR form.
;Such a string is prepended by 4-bytes holding the number of bytes used by the character array (not including the null).
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

  ;The following mask allows us to retrieve the underlying variant type of variables of type #minScript_VARIANT.
    #minScript_VARIANTMASK = $ffff&(~#minScript_VARIANT)
;/////////////////////////////////////////////////////////////////////////////////

;-INCLUDES.
  XIncludeFile #minScript_HOME + "Modules/minScriptSymbols.pbi"

;-DECLARES.

  Declare.i minScript_VarClear(*var.minScriptVariant, blnDoNotClearVTfield=0)

;-STRING FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following function allocates a minScript string and returns the address of the underlying character buffer.
;Returns zero if there is a memory allocation problem.
Procedure.i minScript_StringAlloc(characterBuffer)
  Protected *result.LONG, *ptrChar.CHARACTER, len
  If characterBuffer
    ;We need to calculate the length of the string.
      *ptrChar = characterBuffer
      While *ptrChar\c
        *ptrChar + SizeOf(CHARACTER)
      Wend
      len = *ptrChar - characterBuffer
    ;Allocate required memory.
      *result = AllocateMemory(len + SizeOf(CHARACTER) + SizeOf(LONG))
      If *result
        *result\l = len
        *result + SizeOf(LONG)
        ;Now copy the string.
          CopyMemory(characterBuffer, *result, len)
      EndIf
  EndIf
  ProcedureReturn *result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function allocates a minScript string of the required length (characters) but does not initialise the string.
;Returns the address of the underlying character buffer or zero if there is a memory allocation problem.
Procedure.i minScript_StringAllocLen(numCharacters)
  Protected *result.LONG, lenBytes
  If numCharacters
    lenBytes = numCharacters << (SizeOf(CHARACTER)-1)
    *result = AllocateMemory(lenBytes + SizeOf(CHARACTER) + SizeOf(LONG))
    If *result
      *result\l = lenBytes
      *result + SizeOf(LONG)
    EndIf
  EndIf
  ProcedureReturn *result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function frees a minScript string.
;No return value.
Procedure minScript_StringFree(minScriptString)
  If minScriptString
    FreeMemory(minScriptString-SizeOf(LONG)) 
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VARIANT FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following function coerces the given variable to the specified type.
;If the original variable is marked as a variant then we preserve that.
;Returns a minScript status/error code.
Procedure.i minScript_VarChangeType(*var.minScriptVariant, varType)
  Protected result = #minScript_OKAY, intVal, dblVal.d, t1$, t1, variantBit
  If *var
    ;Record whether the original variable is marked as a variant.
      variantBit = *var\vt&#minscript_VARIANT
    ;Remove any variant flags from the specified type and the given variable.
      varType&#minscript_VARIANTMASK
      *var\vt&#minscript_VARIANTMASK
    If *var\vt <> varType 
      If varType & #minScript_REFERENCE
        ;We only allow a null value to be converted to an object.
        If *var\vt = #minScript_EMPTY Or (*var\vt = #minScript_INTEGER And *var\i = 0)
          *var\vt = varType
        Else
          result = #minScript_INCOMPATIBLEDATATYPES
        EndIf
      Else
        Select varType
          Case #minScript_EMPTY
            result = minScript_VarClear(*var)
          Case #minScript_INTEGER
            If *var\vt & #minScript_REFERENCE
              ;We simply convert to an integer with a value equal to the refPtr interface pointer.
              If *var\refPtr
                intVal = *var\refPtr
                result = *var\refPtr\Clear()
                *var\i = intVal
              EndIf
            Else
              Select *var\vt
                Case #minScript_EMPTY
                  *var\i = 0 ;Just in case.
                Case #minScript_CURRENCY
                  *var\i = (*var\c + 5000) / 10000
                Case #minScript_REAL
                  *var\i = *var\r
                Case #minScript_STRING
                  If *var\sPtr
                    If minScript_SymbolsIsInteger(*var\sPtr, @intVal) = #minScript_OKAY
                      minScript_StringFree(*var\sPtr)
                      *var\i = intVal
                    Else ;Try a real value which will be implicitly cast as an integer (rounded).
                      minScript_SymbolsIsReal(*var\sPtr, @dblVal)
                      minScript_StringFree(*var\sPtr)
                      *var\i = dblVal
                    EndIf
                  Else
                    *var\i = 0
                  EndIf
                Default
                  result = #minScript_INCOMPATIBLEDATATYPES
                EndSelect          
            EndIf
          Case #minScript_REAL
            If *var\vt & #minScript_REFERENCE
              ;We simply convert to a double with a value equal to the refPtr interface pointer.
              If *var\refPtr
                intVal = *var\refPtr
                result = *var\refPtr\Clear()
                *var\r = intVal
              EndIf
            Else
              Select *var\vt
                Case #minScript_EMPTY
                  *var\r = 0 ;Just in case.
                Case #minScript_INTEGER
                  *var\r = *var\i
                Case #minScript_CURRENCY
                  *var\r = *var\c / 10000
                Case #minScript_STRING
                  If *var\sPtr
                    minScript_SymbolsIsReal(*var\sPtr, @dblVal)
                    minScript_StringFree(*var\sPtr)
                    *var\r = dblVal
                  Else
                    *var\r = 0
                  EndIf
                Default
                  result = #minScript_INCOMPATIBLEDATATYPES
              EndSelect
            EndIf
          Case #minScript_CURRENCY
            If *var\vt & #minScript_REFERENCE
              ;We simply convert to a double with a value equal to the refPtr interface pointer.
              If *var\refPtr
                intVal = *var\refPtr
                result = *var\refPtr\Clear()
                *var\c = intVal * 10000
              EndIf
            Else
              Select *var\vt
                Case #minScript_EMPTY
                  *var\c = 0 ;Just in case.
                Case #minScript_INTEGER
                  *var\c = *var\i * 10000
                Case #minScript_REAL
                  *var\c = *var\r * 10000
                Case #minScript_STRING
                  If *var\sPtr
                    t1$ = PeekS(*var\sPtr)
                    minScript_StringFree(*var\sPtr)
                    *var\c = ValC(t1$)
                  Else
                    *var\c = 0
                  EndIf
                Default
                  result = #minScript_INCOMPATIBLEDATATYPES
                EndSelect          
            EndIf
          Case #minScript_STRING
            If *var\vt & #minScript_REFERENCE
              ;We simply convert to a string with a value equal to the refPtr interface pointer.
              If *var\refPtr
                intVal = *var\refPtr
                result = *var\refPtr\Clear()
                t1$ = Str(intVal)
              EndIf
            Else
              Select *var\vt
                Case #minScript_EMPTY ;Do nothing.
                Case #minScript_INTEGER
                  t1$ = Str(*var\i)
                Case #minScript_REAL
                  t1$ = StrD(*var\r)
                Case #minScript_CURRENCY
                  t1$ = StrC(*var\c)
                Default
                  result = #minScript_INCOMPATIBLEDATATYPES
              EndSelect
            EndIf
            If result = #minScript_OKAY And t1$ ; Protects against a #minScript_EMPTY variant.
              t1 = minScript_StringAlloc(@t1$)
              If t1 
                *var\sPtr = t1
              Else
                result = #minScript_OUTOFMEMORY
              EndIf
            Else
              *var\sPtr = 0
            EndIf
          Default
            result = #minScript_INVALIDDATATYPE
        EndSelect
        *var\vt = varType
      EndIf    
      ;Reinstate the variant flag.
        *var\vt|variantBit
    EndIf
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function clears the given variant structure.
;Any non-trivial error code generated will be by a reference type and will result in any execution being halted.
;The variant will be emptied regardless of any errors.
;Set the optional parameter blnDoNotClearVTfield to #minScript_VARDONOTCLEARVTFIELD to leave the VT field intact.
Procedure minScript_VarClear(*var.minScriptVariant, blnDoNotClearVTfield=0)
  Protected result = #minScript_OKAY, vt, t1, i
  If *var
    vt = *var\vt&#minScript_VARIANTMASK
    If vt = #minScript_STRING
      ;Here we must free the string.
        If *var\sPtr
          FreeMemory(*var\sPtr-SizeOf(LONG)) 
        EndIf
    ElseIf vt & #minScript_REFERENCE
      ;We must call the relevant \Clear() method.
        If *var\refPtr
          result = *var\refPtr\Clear()
        EndIf
    EndIf
    *var\q = 0
    If blnDoNotClearVTfield=0
      *var\vt = #minScript_EMPTY
    EndIf
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following function copies the given source variable to the destination one.
;Set the blnFreeDestination parameter to #minScript_VARDONOTFREEDESTINATIONVARIABLE or #minScript_VARFREEDESTINATIONVARIABLE to
;have the contents of the destination variable freed (as appropriate) first.
;If the source variable is of 'variant' type, then the destination is not marked as a variant so only the raw contents are copied.
Procedure.i minScript_VarCopy(*varDest.minScriptVariant, *varSource.minScriptVariant, blnFreeDestination=#minScript_VARFREEDESTINATIONVARIABLE)
  Protected result = #minScript_OKAY, *ptrLong.LONG, lenBytes
  ;Do nothing if the dest and source are the same.
  If *varDest <> *varSource
    If *varDest And *varSource
      If blnFreeDestination = #minScript_VARFREEDESTINATIONVARIABLE
        result = minScript_VarClear(*varDest)
      EndIf
      With *varDest
        \vt = *varSource\vt&#minScript_VARIANTMASK
          If \vt = #minScript_STRING
            If *varSource\sPtr
              ;Fastest way of doing this is to allocate memory manually.
              *ptrLong = *varSource\sPtr - SizeOf(LONG)
              lenBytes = *ptrLong\l
              *ptrLong = AllocateMemory(lenBytes + SizeOf(CHARACTER) + SizeOf(LONG))
              If *ptrLong
                *ptrLong\l = lenBytes
                *ptrLong + SizeOf(LONG)
                ;Now copy the string.
                  CopyMemory(*varSource\sPtr, *ptrLong, lenBytes)
                \sPtr = *ptrLong
              Else
                result = #minScript_OUTOFMEMORY        
              EndIf
            EndIf
          ElseIf \vt & #minScript_REFERENCE
            If *varSource\refPtr
              result = *varSource\refPtr\CopyToVariant(*varDest)
            EndIf
          Else
            \q = *varSource\q
          EndIf
      EndWith
    Else
      result = #minScript_INVALIDARGUMENT  
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VARIANT CREATION FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_VarSetInteger(*var.minScriptVariant, value.i, blnFreeVariableFirst=#True)
  Protected result = #minScript_OKAY
  If *var
    If blnFreeVariableFirst
      result = minScript_VarClear(*var, #True)
    EndIf
    With *var
      \vt = #minScript_INTEGER
      \i = value
    EndWith
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_VarSetCurrency(*var.minScriptVariant, value.q, blnFreeVariableFirst=#True)
  Protected result = #minScript_OKAY
  If *var
    If blnFreeVariableFirst
      result = minScript_VarClear(*var, #True)
    EndIf
    With *var
      \vt = #minScript_CURRENCY
      \q = value
    EndWith
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_VarSetReal(*var.minScriptVariant, value.d, blnFreeVariableFirst=#True)
  Protected result = #minScript_OKAY
  If *var
    If blnFreeVariableFirst
      result = minScript_VarClear(*var, #True)
    EndIf
    With *var
      \vt = #minScript_REAL
      \r = value
    EndWith
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_VarSetString(*var.minScriptVariant, value$, blnFreeVariableFirst=#True)
  Protected result = #minScript_OKAY
  If *var
    If blnFreeVariableFirst
      result = minScript_VarClear(*var, #True)
    EndIf
    With *var
      \vt = #minScript_STRING
      \sPtr = minScript_StringAlloc(@value$)
      If \sPtr = 0 And @value$
        result = #minScript_OUTOFMEMORY
      EndIf
    EndWith
  Else
    result = #minScript_INVALIDARGUMENT  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VARIANT READ FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_VarGetInteger(*var.minScriptVariant)
  If *var
    ProcedureReturn *var\i
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.q minScript_VarGetCurrency(*var.minScriptVariant)
  If *var
    ProcedureReturn *var\c
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.d minScript_VarGetReal(*var.minScriptVariant)
  If *var
    ProcedureReturn *var\r
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScript_VarGetString(*var.minScriptVariant)
  Protected result$
  If *var And *var\sPtr
    result$ = PeekS(*var\sPtr)
  EndIf
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


; IDE Options = PureBasic 5.62 beta 1 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 164
; FirstLine = 139
; Folding = ---
; Executable = d.exe.dll.exe.dll
; CompileSourceDirectory